# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, Dict, Union

import torch

from pyro.poutine.messenger import Messenger
from pyro.poutine.trace_struct import Trace

if TYPE_CHECKING:
    from pyro.poutine.runtime import Message


class ConditionMessenger(Messenger):
    """
    Given a stochastic function with some sample statements
    and a dictionary of observations at names,
    change the sample statements at those names into observes
    with those values.

    Consider the following Pyro program:

        >>> def model(x):
        ...     s = pyro.param("s", torch.tensor(0.5))
        ...     z = pyro.sample("z", dist.Normal(x, s))
        ...     return z ** 2

    To observe a value for site `z`, we can write

        >>> conditioned_model = pyro.poutine.condition(model, data={"z": torch.tensor(1.)})

    This is equivalent to adding `obs=value` as a keyword argument
    to `pyro.sample("z", ...)` in `model`.

    :param fn: a stochastic function (callable containing Pyro primitive calls)
    :param data: a dict or a :class:`~pyro.poutine.Trace`
    :returns: stochastic function decorated with a :class:`~pyro.poutine.condition_messenger.ConditionMessenger`
    """

    def __init__(self, data: Union[Dict[str, torch.Tensor], Trace]) -> None:
        """
        :param data: a dict or a Trace

        Constructor. Doesn't do much, just stores the stochastic function
        and the data to condition on.
        """
        super().__init__()
        self.data = data

    def _pyro_sample(self, msg: "Message") -> None:
        """
        :param msg: current message at a trace site.
        :returns: a sample from the stochastic function at the site.

        If msg["name"] appears in self.data,
        convert the sample site into an observe site
        whose observed value is the value from self.data[msg["name"]].

        Otherwise, implements default sampling behavior
        with no additional effects.
        """
        assert isinstance(msg["name"], str)
        name = msg["name"]

        if name in self.data:
            if isinstance(self.data, Trace):
                msg["value"] = self.data.nodes[name]["value"]
            else:
                msg["value"] = self.data[name]
            msg["is_observed"] = msg["value"] is not None



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

from pyro.distributions.transforms.haar import HaarTransform

from .unit_jacobian import UnitJacobianReparam


class HaarReparam(UnitJacobianReparam):
    """
    Haar wavelet reparameterizer, using a
    :class:`~pyro.distributions.transforms.HaarTransform`.

    This is useful for sequential models where coupling along a time-like axis
    (e.g. a banded precision matrix) introduces long-range correlation. This
    reparameterizes to a frequency-domain representation where posterior
    covariance should be closer to diagonal, thereby improving the accuracy of
    diagonal guides in SVI and improving the effectiveness of a diagonal mass
    matrix in HMC.

    This reparameterization works only for latent variables, not likelihoods.

    :param int dim: Dimension along which to transform. Must be negative.
        This is an absolute dim counting from the right.
    :param bool flip: Whether to flip the time axis before applying the
        Haar transform. Defaults to false.
    :param bool experimental_allow_batch: EXPERIMENTAL allow coupling across a
        batch dimension. The targeted batch dimension and all batch dimensions
        to the right will be converted to event dimensions. Defaults to False.
    """

    def __init__(self, dim=-1, flip=False, *, experimental_allow_batch=False):
        transform = HaarTransform(dim=dim, flip=flip, cache_size=1)
        super().__init__(
            transform, suffix="haar", experimental_allow_batch=experimental_allow_batch
        )



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import torch
import torch.nn as nn
from torch.distributions import MultivariateNormal, constraints

import pyro.distributions as dist
from pyro.contrib.timeseries.base import TimeSeriesModel
from pyro.nn import PyroParam, pyro_method
from pyro.ops.tensor_utils import repeated_matmul


class GenericLGSSM(TimeSeriesModel):
    """
    A generic Linear Gaussian State Space Model parameterized with arbitrary time invariant
    transition and observation dynamics. The targets are (implicitly) assumed to be evenly
    spaced in time. Training and inference are logarithmic in the length of the time series T.

    :param int obs_dim: The dimension of the targets at each time step.
    :param int state_dim: The dimension of latent state at each time step.
    :param bool learnable_observation_loc: whether the mean of the observation model should be learned or not;
        defaults to False.
    """

    def __init__(
        self,
        obs_dim=1,
        state_dim=2,
        obs_noise_scale_init=None,
        learnable_observation_loc=False,
    ):
        self.obs_dim = obs_dim
        self.state_dim = state_dim

        if obs_noise_scale_init is None:
            obs_noise_scale_init = 0.2 * torch.ones(obs_dim)
        assert obs_noise_scale_init.shape == (obs_dim,)

        super().__init__()

        self.obs_noise_scale = PyroParam(
            obs_noise_scale_init, constraint=constraints.positive
        )
        self.trans_noise_scale_sq = PyroParam(
            torch.ones(state_dim), constraint=constraints.positive
        )
        self.trans_matrix = nn.Parameter(
            torch.eye(state_dim) + 0.03 * torch.randn(state_dim, state_dim)
        )
        self.obs_matrix = nn.Parameter(0.3 * torch.randn(state_dim, obs_dim))
        self.init_noise_scale_sq = PyroParam(
            torch.ones(state_dim), constraint=constraints.positive
        )

        if learnable_observation_loc:
            self.obs_loc = nn.Parameter(torch.zeros(obs_dim))
        else:
            self.register_buffer("obs_loc", torch.zeros(obs_dim))

    def _get_init_dist(self):
        loc = self.obs_matrix.new_zeros(self.state_dim)
        return MultivariateNormal(loc, self.init_noise_scale_sq.diag_embed())

    def _get_obs_dist(self):
        return dist.Normal(self.obs_loc, self.obs_noise_scale).to_event(1)

    def _get_trans_dist(self):
        loc = self.obs_matrix.new_zeros(self.state_dim)
        return MultivariateNormal(loc, self.trans_noise_scale_sq.diag_embed())

    def get_dist(self, duration=None):
        """
        Get the :class:`~pyro.distributions.GaussianHMM` distribution that corresponds to :class:`GenericLGSSM`.

        :param int duration: Optional size of the time axis ``event_shape[0]``.
            This is required when sampling from homogeneous HMMs whose parameters
            are not expanded along the time axis.
        """
        return dist.GaussianHMM(
            self._get_init_dist(),
            self.trans_matrix,
            self._get_trans_dist(),
            self.obs_matrix,
            self._get_obs_dist(),
            duration=duration,
        )

    @pyro_method
    def log_prob(self, targets):
        """
        :param torch.Tensor targets: A 2-dimensional tensor of real-valued targets
            of shape ``(T, obs_dim)``, where ``T`` is the length of the time series and ``obs_dim``
            is the dimension of the real-valued ``targets`` at each time step
        :returns torch.Tensor: A (scalar) log probability.
        """
        assert targets.dim() == 2 and targets.size(-1) == self.obs_dim
        return self.get_dist().log_prob(targets)

    @torch.no_grad()
    def _filter(self, targets):
        """
        Return the filtering state for the associated state space model.
        """
        assert targets.dim() == 2 and targets.size(-1) == self.obs_dim
        return self.get_dist().filter(targets)

    @torch.no_grad()
    def _forecast(self, N_timesteps, filtering_state, include_observation_noise=True):
        """
        Internal helper for forecasting.
        """
        N_trans_matrix = repeated_matmul(self.trans_matrix, N_timesteps)
        N_trans_obs = torch.matmul(N_trans_matrix, self.obs_matrix)
        predicted_mean = torch.matmul(filtering_state.loc, N_trans_obs)

        # first compute the contribution from filtering_state.covariance_matrix
        predicted_covar1 = torch.matmul(
            N_trans_obs.transpose(-1, -2),
            torch.matmul(filtering_state.covariance_matrix, N_trans_obs),
        )  # N O O

        # next compute the contribution from process noise that is injected at each timestep.
        # (we need to do a cumulative sum to integrate across time)
        process_covar = self._get_trans_dist().covariance_matrix
        N_trans_obs_shift = torch.cat([self.obs_matrix.unsqueeze(0), N_trans_obs[:-1]])
        predicted_covar2 = torch.matmul(
            N_trans_obs_shift.transpose(-1, -2),
            torch.matmul(process_covar, N_trans_obs_shift),
        )  # N O O

        predicted_covar = predicted_covar1 + torch.cumsum(predicted_covar2, dim=0)

        if include_observation_noise:
            predicted_covar = (
                predicted_covar + self.obs_noise_scale.pow(2.0).diag_embed()
            )

        return predicted_mean, predicted_covar

    @pyro_method
    def forecast(self, targets, N_timesteps):
        """
        :param torch.Tensor targets: A 2-dimensional tensor of real-valued targets
            of shape ``(T, obs_dim)``, where ``T`` is the length of the time series and ``obs_dim``
            is the dimension of the real-valued targets at each time step. These
            represent the training data that are conditioned on for the purpose of making
            forecasts.
        :param int N_timesteps: The number of timesteps to forecast into the future from
            the final target ``targets[-1]``.
        :returns torch.distributions.MultivariateNormal: Returns a predictive MultivariateNormal distribution
            with batch shape ``(N_timesteps,)`` and event shape ``(obs_dim,)``
        """
        filtering_state = self._filter(targets)
        predicted_mean, predicted_covar = self._forecast(N_timesteps, filtering_state)
        return torch.distributions.MultivariateNormal(predicted_mean, predicted_covar)



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import warnings

import torch

from pyro.infer.trace_elbo import Trace_ELBO
from pyro.infer.util import check_fully_reparametrized, is_validation_enabled


class TraceTailAdaptive_ELBO(Trace_ELBO):
    """
    Interface for Stochastic Variational Inference with an adaptive
    f-divergence as described in ref. [1]. Users should specify
    `num_particles` > 1 and `vectorize_particles==True`. The argument
    `tail_adaptive_beta` can be specified to modify how the adaptive
    f-divergence is constructed. See reference for details.

    Note that this interface does not support computing the varational
    objective itself; rather it only supports computing gradients of the
    variational objective. Consequently, one might want to use
    another SVI interface (e.g. `RenyiELBO`) in order to monitor convergence.

    Note that this interface only supports models in which all the latent
    variables are fully reparameterized. It also does not support data
    subsampling.

    References
    [1] "Variational Inference with Tail-adaptive f-Divergence", Dilin Wang,
    Hao Liu, Qiang Liu, NeurIPS 2018
    https://papers.nips.cc/paper/7816-variational-inference-with-tail-adaptive-f-divergence
    """

    def loss(self, model, guide, *args, **kwargs):
        """
        It is not necessary to estimate the tail-adaptive f-divergence itself in order
        to compute the corresponding gradients. Consequently the loss method is left
        unimplemented.
        """
        raise NotImplementedError(
            "Loss method for TraceTailAdaptive_ELBO not implemented"
        )

    def _differentiable_loss_particle(self, model_trace, guide_trace):
        if not self.vectorize_particles:
            raise NotImplementedError(
                "TraceTailAdaptive_ELBO only implemented for vectorize_particles==True"
            )

        if self.num_particles == 1:
            warnings.warn(
                "For num_particles==1 TraceTailAdaptive_ELBO uses the same loss function as Trace_ELBO. "
                + "Increase num_particles to get an adaptive f-divergence."
            )

        log_p, log_q = 0, 0

        for name, site in model_trace.nodes.items():
            if site["type"] == "sample":
                site_log_p = site["log_prob"].reshape(self.num_particles, -1).sum(-1)
                log_p = log_p + site_log_p

        for name, site in guide_trace.nodes.items():
            if site["type"] == "sample":
                site_log_q = site["log_prob"].reshape(self.num_particles, -1).sum(-1)
                log_q = log_q + site_log_q
                if is_validation_enabled():
                    check_fully_reparametrized(site)

        # rank the particles according to p/q
        log_pq = log_p - log_q
        rank = torch.argsort(log_pq, descending=False)
        rank = torch.index_select(
            torch.arange(self.num_particles, device=log_pq.device) + 1, -1, rank
        ).type_as(log_pq)

        # compute the particle-specific weights used to construct the surrogate loss
        gamma = torch.pow(rank, self.tail_adaptive_beta).detach()
        surrogate_loss = -(log_pq * gamma).sum() / gamma.sum()

        # we do not compute the loss, so return `inf`
        return float("inf"), surrogate_loss



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

from collections import Counter, OrderedDict, namedtuple
from enum import Enum


class StackFrame:
    """
    Consistent bidirectional mapping between integer positional dimensions and names.
    Can be queried like a dictionary (``value = frame[key]``, ``frame[key] = value``).
    """

    def __init__(self, name_to_dim, dim_to_name, history=1, keep=False):
        assert isinstance(name_to_dim, OrderedDict) and all(
            isinstance(name, str) and isinstance(dim, int)
            for name, dim in name_to_dim.items()
        )
        assert isinstance(dim_to_name, OrderedDict) and all(
            isinstance(name, str) and isinstance(dim, int)
            for dim, name in dim_to_name.items()
        )
        self.name_to_dim = name_to_dim
        self.dim_to_name = dim_to_name
        self.history = history
        self.keep = keep

    def __setitem__(self, key, value):
        assert (
            isinstance(key, (int, str))
            and isinstance(value, (int, str))
            and type(key) != type(value)
        )
        name, dim = (value, key) if isinstance(key, int) else (key, value)
        self.name_to_dim[name], self.dim_to_name[dim] = dim, name

    def __getitem__(self, key):
        assert isinstance(key, (int, str))
        return self.dim_to_name[key] if isinstance(key, int) else self.name_to_dim[key]

    def __delitem__(self, key):
        assert isinstance(key, (int, str))
        k2v, v2k = (
            (self.dim_to_name, self.name_to_dim)
            if isinstance(key, int)
            else (self.name_to_dim, self.dim_to_name)
        )
        del v2k[k2v[key]]
        del k2v[key]

    def __contains__(self, key):
        assert isinstance(key, (int, str))
        return key in (self.dim_to_name if isinstance(key, int) else self.name_to_dim)


class DimType(Enum):
    """Enumerates the possible types of dimensions to allocate"""

    LOCAL = 0
    GLOBAL = 1
    VISIBLE = 2


DimRequest = namedtuple("DimRequest", ["value", "dim_type"])
DimRequest.__new__.__defaults__ = (None, DimType.LOCAL)


class DimStack:
    """
    Single piece of global state to keep track of the mapping between names and dimensions.

    Replaces the plate :class:`~pyro.poutine.runtime._DimAllocator`,
    the enum :class:`~pyro.poutine.runtime._EnumAllocator`, the ``stack`` in :class:`~MarkovMessenger`,
    ``_param_dims`` and ``_value_dims`` in :class:`~EnumMessenger`, and ``dim_to_symbol`` in ``msg['infer']``
    """

    def __init__(self):
        global_frame = StackFrame(
            name_to_dim=OrderedDict(),
            dim_to_name=OrderedDict(),
            history=0,
            keep=False,
        )
        self._local_stack = [global_frame]
        self._iter_stack = [global_frame]
        self._global_stack = [global_frame]
        self._first_available_dim = self.DEFAULT_FIRST_DIM
        self.outermost = None

    MAX_DIM = -25
    DEFAULT_FIRST_DIM = -5

    def set_first_available_dim(self, dim):
        assert dim is None or (self.MAX_DIM < dim < 0)
        old_dim, self._first_available_dim = self._first_available_dim, dim
        return old_dim

    def push_global(self, frame):
        self._global_stack.append(frame)

    def pop_global(self):
        assert self._global_stack, "cannot pop the global frame"
        return self._global_stack.pop()

    def push_iter(self, frame):
        self._iter_stack.append(frame)

    def pop_iter(self):
        assert self._iter_stack, "cannot pop the global frame"
        return self._iter_stack.pop()

    def push_local(self, frame):
        self._local_stack.append(frame)

    def pop_local(self):
        assert self._local_stack, "cannot pop the global frame"
        return self._local_stack.pop()

    @property
    def global_frame(self):
        return self._global_stack[-1]

    @property
    def local_frame(self):
        return self._local_stack[-1]

    @property
    def current_write_env(self):
        return (
            self._local_stack[-1:]
            if not self.local_frame.keep
            else self._local_stack[-self.local_frame.history - 1 :]
        )

    @property
    def current_read_env(self):
        """
        Collect all frames necessary to compute the full name <--> dim mapping
        and interpret Funsor inputs or batch shapes at any point in a computation.
        """
        return (
            self._global_stack
            + self._local_stack[-self.local_frame.history - 1 :]
            + self._iter_stack
        )

    def _genvalue(self, key, value_request):
        """
        Given proposed values for a fresh (name, dim) pair, computes a new, possibly
        identical (name, dim) pair consistent with the current name <--> dim mapping.
        This function is pure and does not update the name <--> dim mapping itself.

        The implementation here is only one of several possibilities, and was chosen
        to match the behavior of Pyro's old enumeration machinery as closely as possible.
        """
        if isinstance(key, int):
            dim, name = key, value_request.value
            fresh_name = "_pyro_dim_{}".format(-key) if name is None else name
            return dim, fresh_name

        elif isinstance(key, str):
            name, dim, dim_type = key, value_request.value, value_request.dim_type
            if dim_type == DimType.VISIBLE:
                fresh_dim = -1 if dim is None else dim
            else:
                fresh_dim = self._first_available_dim  # discard input...

            while any(fresh_dim in p for p in self.current_read_env):
                fresh_dim -= 1

            if fresh_dim < self.MAX_DIM or (
                dim_type == DimType.VISIBLE and fresh_dim <= self._first_available_dim
            ):
                raise ValueError(
                    "Ran out of free dims during allocation for {}".format(name)
                )

            return name, fresh_dim
        raise ValueError(
            "{} and {} not a valid name-dim pair".format(key, value_request)
        )

    def allocate(self, key_to_value_request):
        # step 1: split into fresh and non-fresh
        key_to_value = OrderedDict()
        for key, value_request in tuple(key_to_value_request.items()):
            value = value_request.value
            for frame in self.current_read_env:
                if value is None and key in frame:
                    key_to_value[key] = frame[key]
                    del key_to_value_request[key]
                    break
                elif value is not None and value in frame:
                    key_to_value[key] = value
                    del key_to_value_request[key]
                    break

        # step 2: check that the non-fresh input mapping from keys to values is 1-1
        if max(Counter(key_to_value.values()).values(), default=0) > 1:
            raise ValueError("{} is not a valid shape request".format(key_to_value))

        # step 3: allocate fresh values for all fresh
        for key, value_request in key_to_value_request.items():
            key, fresh_value = self._genvalue(key, value_request)
            # if this key is already active but inconsistent with the fresh value,
            # generate a fresh_key for future conversions via _genvalue in reverse
            if value_request.dim_type != DimType.VISIBLE or any(
                key in frame for frame in self.current_read_env
            ):
                _, fresh_key = self._genvalue(
                    fresh_value, DimRequest(key, value_request.dim_type)
                )
            else:
                fresh_key = key
            for frame in (
                [self.global_frame]
                if value_request.dim_type != DimType.LOCAL
                else self.current_write_env
            ):
                frame[fresh_key] = fresh_value
            # use the user-provided key rather than fresh_key for satisfying this request only
            key_to_value[key] = fresh_value

        assert not any(isinstance(value, DimRequest) for value in key_to_value.values())
        return key_to_value

    def names_from_batch_shape(self, batch_shape, dim_type=DimType.LOCAL):
        return self.allocate_dim_to_name(
            OrderedDict(
                (dim, DimRequest(None, dim_type))
                for dim in range(-len(batch_shape), 0)
                if batch_shape[dim] > 1
            )
        )


_DIM_STACK = DimStack()  # only one global instance



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import operator
from functools import partial, reduce

import torch
from torch.distributions.utils import _sum_rightmost

from pyro.nn import ConditionalDenseNN, DenseNN

from .. import constraints
from ..conditional import ConditionalTransformModule
from ..torch_transform import TransformModule
from ..transforms.utils import clamp_preserve_gradients
from ..util import copy_docs_from


@copy_docs_from(TransformModule)
class AffineCoupling(TransformModule):
    r"""
    An implementation of the affine coupling layer of RealNVP (Dinh et al., 2017)
    that uses the bijective transform,

        :math:`\mathbf{y}_{1:d} = \mathbf{x}_{1:d}`
        :math:`\mathbf{y}_{(d+1):D} = \mu + \sigma\odot\mathbf{x}_{(d+1):D}`

    where :math:`\mathbf{x}` are the inputs, :math:`\mathbf{y}` are the outputs,
    e.g. :math:`\mathbf{x}_{1:d}` represents the first :math:`d` elements of the
    inputs, and :math:`\mu,\sigma` are shift and translation parameters calculated
    as the output of a function inputting only :math:`\mathbf{x}_{1:d}`.

    That is, the first :math:`d` components remain unchanged, and the subsequent
    :math:`D-d` are shifted and translated by a function of the previous components.

    Together with :class:`~pyro.distributions.TransformedDistribution` this provides
    a way to create richer variational approximations.

    Example usage:

    >>> from pyro.nn import DenseNN
    >>> input_dim = 10
    >>> split_dim = 6
    >>> base_dist = dist.Normal(torch.zeros(input_dim), torch.ones(input_dim))
    >>> param_dims = [input_dim-split_dim, input_dim-split_dim]
    >>> hypernet = DenseNN(split_dim, [10*input_dim], param_dims)
    >>> transform = AffineCoupling(split_dim, hypernet)
    >>> pyro.module("my_transform", transform)  # doctest: +SKIP
    >>> flow_dist = dist.TransformedDistribution(base_dist, [transform])
    >>> flow_dist.sample()  # doctest: +SKIP

    The inverse of the Bijector is required when, e.g., scoring the log density of a
    sample with :class:`~pyro.distributions.TransformedDistribution`. This
    implementation caches the inverse of the Bijector when its forward operation is
    called, e.g., when sampling from
    :class:`~pyro.distributions.TransformedDistribution`. However, if the cached
    value isn't available, either because it was overwritten during sampling a new
    value or an arbitary value is being scored, it will calculate it manually.

    This is an operation that scales as O(1), i.e. constant in the input dimension.
    So in general, it is cheap to sample *and* score (an arbitrary value) from
    :class:`~pyro.distributions.transforms.AffineCoupling`.

    :param split_dim: Zero-indexed dimension :math:`d` upon which to perform input/
        output split for transformation.
    :type split_dim: int
    :param hypernet: a neural network whose forward call returns a real-valued mean
        and logit-scale as a tuple. The input should have final dimension split_dim
        and the output final dimension input_dim-split_dim for each member of the
        tuple.
    :type hypernet: callable
    :param dim: the tensor dimension on which to split. This value must be negative
        and defines the event dim as `abs(dim)`.
    :type dim: int
    :param log_scale_min_clip: The minimum value for clipping the log(scale) from
        the autoregressive NN
    :type log_scale_min_clip: float
    :param log_scale_max_clip: The maximum value for clipping the log(scale) from
        the autoregressive NN
    :type log_scale_max_clip: float

    References:

    [1] Laurent Dinh, Jascha Sohl-Dickstein, and Samy Bengio. Density estimation
    using Real NVP. ICLR 2017.

    """

    bijective = True

    def __init__(
        self,
        split_dim,
        hypernet,
        *,
        dim=-1,
        log_scale_min_clip=-5.0,
        log_scale_max_clip=3.0
    ):
        super().__init__(cache_size=1)
        if dim >= 0:
            raise ValueError("'dim' keyword argument must be negative")

        self.split_dim = split_dim
        self.nn = hypernet
        self.dim = dim
        self._cached_log_scale = None
        self.log_scale_min_clip = log_scale_min_clip
        self.log_scale_max_clip = log_scale_max_clip

    @constraints.dependent_property(is_discrete=False)
    def domain(self):
        return constraints.independent(constraints.real, -self.dim)

    @constraints.dependent_property(is_discrete=False)
    def codomain(self):
        return constraints.independent(constraints.real, -self.dim)

    def _call(self, x):
        """
        :param x: the input into the bijection
        :type x: torch.Tensor

        Invokes the bijection x=>y; in the prototypical context of a
        :class:`~pyro.distributions.TransformedDistribution` `x` is a sample from
        the base distribution (or the output of a previous transform)
        """
        x1, x2 = x.split(
            [self.split_dim, x.size(self.dim) - self.split_dim], dim=self.dim
        )

        # Now that we can split on an arbitrary dimension, we have do a bit of reshaping...
        mean, log_scale = self.nn(x1.reshape(x1.shape[: self.dim] + (-1,)))
        mean = mean.reshape(mean.shape[:-1] + x2.shape[self.dim :])
        log_scale = log_scale.reshape(log_scale.shape[:-1] + x2.shape[self.dim :])

        log_scale = clamp_preserve_gradients(
            log_scale, self.log_scale_min_clip, self.log_scale_max_clip
        )
        self._cached_log_scale = log_scale

        y1 = x1
        y2 = torch.exp(log_scale) * x2 + mean
        return torch.cat([y1, y2], dim=self.dim)

    def _inverse(self, y):
        """
        :param y: the output of the bijection
        :type y: torch.Tensor

        Inverts y => x. Uses a previously cached inverse if available, otherwise
        performs the inversion afresh.
        """
        y1, y2 = y.split(
            [self.split_dim, y.size(self.dim) - self.split_dim], dim=self.dim
        )
        x1 = y1

        # Now that we can split on an arbitrary dimension, we have do a bit of reshaping...
        mean, log_scale = self.nn(x1.reshape(x1.shape[: self.dim] + (-1,)))
        mean = mean.reshape(mean.shape[:-1] + y2.shape[self.dim :])
        log_scale = log_scale.reshape(log_scale.shape[:-1] + y2.shape[self.dim :])

        log_scale = clamp_preserve_gradients(
            log_scale, self.log_scale_min_clip, self.log_scale_max_clip
        )
        self._cached_log_scale = log_scale

        x2 = (y2 - mean) * torch.exp(-log_scale)
        return torch.cat([x1, x2], dim=self.dim)

    def log_abs_det_jacobian(self, x, y):
        """
        Calculates the elementwise determinant of the log jacobian
        """
        x_old, y_old = self._cached_x_y
        if self._cached_log_scale is not None and x is x_old and y is y_old:
            log_scale = self._cached_log_scale
        else:
            x1, x2 = x.split(
                [self.split_dim, x.size(self.dim) - self.split_dim], dim=self.dim
            )
            _, log_scale = self.nn(x1.reshape(x1.shape[: self.dim] + (-1,)))
            log_scale = log_scale.reshape(log_scale.shape[:-1] + x2.shape[self.dim :])
            log_scale = clamp_preserve_gradients(
                log_scale, self.log_scale_min_clip, self.log_scale_max_clip
            )
        return _sum_rightmost(log_scale, self.event_dim)


@copy_docs_from(ConditionalTransformModule)
class ConditionalAffineCoupling(ConditionalTransformModule):
    r"""
    An implementation of the affine coupling layer of RealNVP (Dinh et al., 2017)
    that conditions on an additional context variable and uses the bijective
    transform,

        :math:`\mathbf{y}_{1:d} = \mathbf{x}_{1:d}`
        :math:`\mathbf{y}_{(d+1):D} = \mu + \sigma\odot\mathbf{x}_{(d+1):D}`

    where :math:`\mathbf{x}` are the inputs, :math:`\mathbf{y}` are the outputs,
    e.g. :math:`\mathbf{x}_{1:d}` represents the first :math:`d` elements of the
    inputs, and :math:`\mu,\sigma` are shift and translation parameters calculated
    as the output of a function input :math:`\mathbf{x}_{1:d}` and a context
    variable :math:`\mathbf{z}\in\mathbb{R}^M`.

    That is, the first :math:`d` components remain unchanged, and the subsequent
    :math:`D-d` are shifted and translated by a function of the previous components.

    Together with :class:`~pyro.distributions.ConditionalTransformedDistribution`
    this provides a way to create richer variational approximations.

    Example usage:

    >>> from pyro.nn import ConditionalDenseNN
    >>> input_dim = 10
    >>> split_dim = 6
    >>> context_dim = 4
    >>> batch_size = 3
    >>> base_dist = dist.Normal(torch.zeros(input_dim), torch.ones(input_dim))
    >>> param_dims = [input_dim-split_dim, input_dim-split_dim]
    >>> hypernet = ConditionalDenseNN(split_dim, context_dim, [10*input_dim],
    ... param_dims)
    >>> transform = ConditionalAffineCoupling(split_dim, hypernet)
    >>> pyro.module("my_transform", transform)  # doctest: +SKIP
    >>> z = torch.rand(batch_size, context_dim)
    >>> flow_dist = dist.ConditionalTransformedDistribution(base_dist,
    ... [transform]).condition(z)
    >>> flow_dist.sample(sample_shape=torch.Size([batch_size]))  # doctest: +SKIP

    The inverse of the Bijector is required when, e.g., scoring the log density of a
    sample with :class:`~pyro.distributions.ConditionalTransformedDistribution`.
    This implementation caches the inverse of the Bijector when its forward
    operation is called, e.g., when sampling from
    :class:`~pyro.distributions.ConditionalTransformedDistribution`. However, if the
    cached value isn't available, either because it was overwritten during sampling
    a new value or an arbitary value is being scored, it will calculate it manually.

    This is an operation that scales as O(1), i.e. constant in the input dimension.
    So in general, it is cheap to sample *and* score (an arbitrary value) from
    :class:`~pyro.distributions.transforms.ConditionalAffineCoupling`.

    :param split_dim: Zero-indexed dimension :math:`d` upon which to perform input/
        output split for transformation.
    :type split_dim: int
    :param hypernet: A neural network whose forward call returns a real-valued mean
        and logit-scale as a tuple. The input should have final dimension split_dim
        and the output final dimension input_dim-split_dim for each member of the
        tuple. The network also inputs a context variable as a keyword argument in
        order to condition the output upon it.
    :type hypernet: callable
    :param log_scale_min_clip: The minimum value for clipping the log(scale) from
        the NN
    :type log_scale_min_clip: float
    :param log_scale_max_clip: The maximum value for clipping the log(scale) from
        the NN
    :type log_scale_max_clip: float

    References:

    Laurent Dinh, Jascha Sohl-Dickstein, and Samy Bengio. Density estimation using
    Real NVP. ICLR 2017.

    """

    domain = constraints.real_vector
    codomain = constraints.real_vector
    bijective = True

    def __init__(self, split_dim, hypernet, **kwargs):
        super().__init__()
        self.split_dim = split_dim
        self.nn = hypernet
        self.kwargs = kwargs

    def condition(self, context):
        cond_nn = partial(self.nn, context=context)
        return AffineCoupling(self.split_dim, cond_nn, **self.kwargs)


def affine_coupling(input_dim, hidden_dims=None, split_dim=None, dim=-1, **kwargs):
    """
    A helper function to create an
    :class:`~pyro.distributions.transforms.AffineCoupling` object that takes care of
    constructing a dense network with the correct input/output dimensions.

    :param input_dim: Dimension(s) of input variable to permute. Note that when
        `dim < -1` this must be a tuple corresponding to the event shape.
    :type input_dim: int
    :param hidden_dims: The desired hidden dimensions of the dense network. Defaults
        to using [10*input_dim]
    :type hidden_dims: list[int]
    :param split_dim: The dimension to split the input on for the coupling
        transform. Defaults to using input_dim // 2
    :type split_dim: int
    :param dim: the tensor dimension on which to split. This value must be negative
        and defines the event dim as `abs(dim)`.
    :type dim: int
    :param log_scale_min_clip: The minimum value for clipping the log(scale) from
        the autoregressive NN
    :type log_scale_min_clip: float
    :param log_scale_max_clip: The maximum value for clipping the log(scale) from
        the autoregressive NN
    :type log_scale_max_clip: float

    """
    if not isinstance(input_dim, int):
        if len(input_dim) != -dim:
            raise ValueError(
                "event shape {} must have same length as event_dim {}".format(
                    input_dim, -dim
                )
            )
        event_shape = input_dim
        extra_dims = reduce(operator.mul, event_shape[(dim + 1) :], 1)
    else:
        event_shape = [input_dim]
        extra_dims = 1
    event_shape = list(event_shape)

    if split_dim is None:
        split_dim = event_shape[dim] // 2
    if hidden_dims is None:
        hidden_dims = [10 * event_shape[dim] * extra_dims]

    hypernet = DenseNN(
        split_dim * extra_dims,
        hidden_dims,
        [
            (event_shape[dim] - split_dim) * extra_dims,
            (event_shape[dim] - split_dim) * extra_dims,
        ],
    )
    return AffineCoupling(split_dim, hypernet, dim=dim, **kwargs)


def conditional_affine_coupling(
    input_dim, context_dim, hidden_dims=None, split_dim=None, dim=-1, **kwargs
):
    """
    A helper function to create an
    :class:`~pyro.distributions.transforms.ConditionalAffineCoupling` object that
    takes care of constructing a dense network with the correct input/output
    dimensions.

    :param input_dim: Dimension of input variable
    :type input_dim: int
    :param context_dim: Dimension of context variable
    :type context_dim: int
    :param hidden_dims: The desired hidden dimensions of the dense network. Defaults
        to using [10*input_dim]
    :type hidden_dims: list[int]
    :param split_dim: The dimension to split the input on for the coupling
        transform. Defaults to using input_dim // 2
    :type split_dim: int
    :param dim: the tensor dimension on which to split. This value must be negative
        and defines the event dim as `abs(dim)`.
    :type dim: int
    :param log_scale_min_clip: The minimum value for clipping the log(scale) from
        the autoregressive NN
    :type log_scale_min_clip: float
    :param log_scale_max_clip: The maximum value for clipping the log(scale) from
        the autoregressive NN
    :type log_scale_max_clip: float

    """
    if not isinstance(input_dim, int):
        if len(input_dim) != -dim:
            raise ValueError(
                "event shape {} must have same length as event_dim {}".format(
                    input_dim, -dim
                )
            )
        event_shape = input_dim
        extra_dims = reduce(operator.mul, event_shape[(dim + 1) :], 1)
    else:
        event_shape = [input_dim]
        extra_dims = 1
    event_shape = list(event_shape)

    if split_dim is None:
        split_dim = event_shape[dim] // 2
    if hidden_dims is None:
        hidden_dims = [10 * event_shape[dim] * extra_dims]

    nn = ConditionalDenseNN(
        split_dim * extra_dims,
        context_dim,
        hidden_dims,
        [
            (event_shape[dim] - split_dim) * extra_dims,
            (event_shape[dim] - split_dim) * extra_dims,
        ],
    )
    return ConditionalAffineCoupling(split_dim, nn, dim=dim, **kwargs)



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import torch

from pyro.distributions.score_parts import ScoreParts
from pyro.distributions.torch_distribution import TorchDistribution


class Rejector(TorchDistribution):
    """
    Rejection sampled distribution given an acceptance rate function.

    :param Distribution propose: A proposal distribution that samples batched
        proposals via ``propose()``. :meth:`rsample` supports a ``sample_shape``
        arg only if ``propose()`` supports a ``sample_shape`` arg.
    :param callable log_prob_accept: A callable that inputs a batch of
        proposals and returns a batch of log acceptance probabilities.
    :param log_scale: Total log probability of acceptance.
    """

    arg_constraints = {}
    has_rsample = True

    def __init__(
        self, propose, log_prob_accept, log_scale, *, batch_shape=None, event_shape=None
    ):
        self.propose = propose
        self.log_prob_accept = log_prob_accept
        self._log_scale = log_scale
        if batch_shape is None:
            batch_shape = propose.batch_shape
        if event_shape is None:
            event_shape = propose.event_shape
        super().__init__(batch_shape, event_shape)

        # These LRU(1) caches allow work to be shared across different method calls.
        self._log_prob_accept_cache = None, None
        self._propose_log_prob_cache = None, None

    def _log_prob_accept(self, x):
        if x is not self._log_prob_accept_cache[0]:
            self._log_prob_accept_cache = x, self.log_prob_accept(x) - self._log_scale
        return self._log_prob_accept_cache[1]

    def _propose_log_prob(self, x):
        if x is not self._propose_log_prob_cache[0]:
            self._propose_log_prob_cache = x, self.propose.log_prob(x)
        return self._propose_log_prob_cache[1]

    def rsample(self, sample_shape=torch.Size()):
        # Implements parallel batched accept-reject sampling.
        x = self.propose(sample_shape) if sample_shape else self.propose()
        log_prob_accept = self.log_prob_accept(x)
        probs = torch.exp(log_prob_accept).clamp_(0.0, 1.0)
        done = torch.bernoulli(probs).bool()
        while not done.all():
            proposed_x = self.propose(sample_shape) if sample_shape else self.propose()
            log_prob_accept = self.log_prob_accept(proposed_x)
            prob_accept = torch.exp(log_prob_accept).clamp_(0.0, 1.0)
            accept = torch.bernoulli(prob_accept).bool() & ~done
            if accept.any():
                x[accept] = proposed_x[accept]
                done |= accept
        return x

    def log_prob(self, x):
        return self._propose_log_prob(x) + self._log_prob_accept(x)

    def score_parts(self, x):
        score_function = self._log_prob_accept(x)
        log_prob = self.log_prob(x)
        return ScoreParts(log_prob, score_function, log_prob)



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

import torch
from torch.distributions import constraints
from torch.distributions.utils import broadcast_all, lazy_property

from pyro.distributions.torch import NegativeBinomial
from pyro.distributions.torch_distribution import TorchDistribution
from pyro.distributions.util import broadcast_shape
from pyro.ops.special import get_quad_rule


class LogNormalNegativeBinomial(TorchDistribution):
    r"""
    A three-parameter generalization of the Negative Binomial distribution [1].
    It can be understood as a continuous mixture of Negative Binomial distributions
    in which we inject Normally-distributed noise into the logits of the Negative
    Binomial distribution:

    .. math::

        \begin{eqnarray}
        &\rm{LNNB}(y | \rm{total\_count}=\nu, \rm{logits}=\ell, \rm{multiplicative\_noise\_scale}=sigma) = \\
        &\int d\epsilon \mathcal{N}(\epsilon | 0, \sigma)
        \rm{NB}(y | \rm{total\_count}=\nu, \rm{logits}=\ell + \epsilon)
        \end{eqnarray}

    where :math:`y \ge 0` is a non-negative integer. Thus while a Negative Binomial distribution
    can be formulated as a Poisson distribution with a Gamma-distributed rate, this distribution
    adds an additional level of variability by also modulating the rate by Log Normally-distributed
    multiplicative noise.

    This distribution has a mean given by

    .. math::
        \mathbb{E}[y] = \nu e^{\ell} = e^{\ell + \log \nu + \tfrac{1}{2}\sigma^2}

    and a variance given by

    .. math::
        \rm{Var}[y] = \mathbb{E}[y] + \left( e^{\sigma^2} (1 + 1/\nu) - 1 \right) \left( \mathbb{E}[y] \right)^2

    Thus while a given mean and variance together uniquely characterize a Negative Binomial distribution, there is a
    one-dimensional family of Log Normal Negative Binomial distributions with a given mean and variance.

    Note that in some applications it may be useful to parameterize the logits as

    .. math::
        \ell = \ell^\prime - \log \nu - \tfrac{1}{2}\sigma^2

    so that the mean is given by :math:`\mathbb{E}[y] = e^{\ell^\prime}` and does not depend on :math:`\nu`
    and :math:`\sigma`, which serve to determine the higher moments.

    References:

    [1] "Lognormal and Gamma Mixed Negative Binomial Regression,"
    Mingyuan Zhou, Lingbo Li, David Dunson, and Lawrence Carin.

    :param total_count: non-negative number of negative Bernoulli trials. The variance decreases
        as `total_count` increases.
    :type total_count: float or torch.Tensor
    :param torch.Tensor logits: Event log-odds for probabilities of success for underlying
        Negative Binomial distribution.
    :param torch.Tensor multiplicative_noise_scale: Controls the level of the injected Normal logit noise.
    :param int num_quad_points: Number of quadrature points used to compute the (approximate) `log_prob`.
        Defaults to 8.
    """

    arg_constraints = {
        "total_count": constraints.greater_than_eq(0),
        "logits": constraints.real,
        "multiplicative_noise_scale": constraints.positive,
    }
    support = constraints.nonnegative_integer

    def __init__(
        self,
        total_count,
        logits,
        multiplicative_noise_scale,
        *,
        num_quad_points=8,
        validate_args=None,
    ):
        if num_quad_points < 1:
            raise ValueError("num_quad_points must be positive.")

        total_count, logits, multiplicative_noise_scale = broadcast_all(
            total_count, logits, multiplicative_noise_scale
        )

        self.quad_points, self.log_weights = get_quad_rule(num_quad_points, logits)
        quad_logits = (
            logits.unsqueeze(-1)
            + multiplicative_noise_scale.unsqueeze(-1) * self.quad_points
        )
        self.nb_dist = NegativeBinomial(
            total_count=total_count.unsqueeze(-1), logits=quad_logits
        )

        self.multiplicative_noise_scale = multiplicative_noise_scale
        self.total_count = total_count
        self.logits = logits
        self.num_quad_points = num_quad_points

        batch_shape = broadcast_shape(
            multiplicative_noise_scale.shape, self.nb_dist.batch_shape[:-1]
        )
        event_shape = torch.Size()

        super().__init__(batch_shape, event_shape, validate_args)

    def log_prob(self, value):
        nb_log_prob = self.nb_dist.log_prob(value.unsqueeze(-1))
        return torch.logsumexp(self.log_weights + nb_log_prob, axis=-1)

    def sample(self, sample_shape=torch.Size()):
        raise NotImplementedError

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(type(self), _instance)
        batch_shape = torch.Size(batch_shape)
        total_count = self.total_count.expand(batch_shape)
        logits = self.logits.expand(batch_shape)
        multiplicative_noise_scale = self.multiplicative_noise_scale.expand(batch_shape)
        LogNormalNegativeBinomial.__init__(
            new,
            total_count,
            logits,
            multiplicative_noise_scale,
            num_quad_points=self.num_quad_points,
            validate_args=False,
        )
        new._validate_args = self._validate_args
        return new

    @lazy_property
    def mean(self):
        return torch.exp(
            self.logits
            + self.total_count.log()
            + 0.5 * self.multiplicative_noise_scale.pow(2.0)
        )

    @lazy_property
    def variance(self):
        kappa = (
            torch.exp(self.multiplicative_noise_scale.pow(2.0))
            * (1 + 1 / self.total_count)
            - 1
        )
        return self.mean + kappa * self.mean.pow(2.0)



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import weakref
from collections import defaultdict
from operator import itemgetter

import torch

import pyro
import pyro.ops.jit
from pyro.distributions.util import detach, is_identically_zero
from pyro.infer import ELBO
from pyro.infer.enum import get_importance_trace
from pyro.infer.util import (
    MultiFrameTensor,
    get_plate_stacks,
    is_validation_enabled,
    torch_backward,
    torch_item,
)
from pyro.ops.provenance import detach_provenance, get_provenance, track_provenance
from pyro.poutine.messenger import Messenger
from pyro.poutine.subsample_messenger import _Subsample
from pyro.util import check_if_enumerated, warn_if_nan


def _get_baseline_options(site):
    """
    Extracts baseline options from ``site["infer"]["baseline"]``.
    """
    # XXX default for baseline_beta currently set here
    options_dict = site["infer"].get("baseline", {}).copy()
    options_tuple = (
        options_dict.pop("nn_baseline", None),
        options_dict.pop("nn_baseline_input", None),
        options_dict.pop("use_decaying_avg_baseline", False),
        options_dict.pop("baseline_beta", 0.90),
        options_dict.pop("baseline_value", None),
    )
    if options_dict:
        raise ValueError(
            "Unrecognized baseline options: {}".format(options_dict.keys())
        )
    return options_tuple


def _construct_baseline(node, guide_site, downstream_cost):
    # XXX should the average baseline be in the param store as below?

    baseline = 0.0
    baseline_loss = 0.0

    (
        nn_baseline,
        nn_baseline_input,
        use_decaying_avg_baseline,
        baseline_beta,
        baseline_value,
    ) = _get_baseline_options(guide_site)

    use_nn_baseline = nn_baseline is not None
    use_baseline_value = baseline_value is not None

    use_baseline = use_nn_baseline or use_decaying_avg_baseline or use_baseline_value

    assert not (
        use_nn_baseline and use_baseline_value
    ), "cannot use baseline_value and nn_baseline simultaneously"
    if use_decaying_avg_baseline:
        dc_shape = downstream_cost.shape
        param_name = "__baseline_avg_downstream_cost_" + node
        with torch.no_grad():
            avg_downstream_cost_old = pyro.param(
                param_name, torch.zeros(dc_shape, device=guide_site["value"].device)
            )
            avg_downstream_cost_new = (
                1 - baseline_beta
            ) * downstream_cost + baseline_beta * avg_downstream_cost_old
        pyro.get_param_store()[param_name] = avg_downstream_cost_new
        baseline += avg_downstream_cost_old
    if use_nn_baseline:
        # block nn_baseline_input gradients except in baseline loss
        baseline += nn_baseline(detach(nn_baseline_input))
    elif use_baseline_value:
        # it's on the user to make sure baseline_value tape only points to baseline params
        baseline += baseline_value
    if use_nn_baseline or use_baseline_value:
        # accumulate baseline loss
        baseline_loss += torch.pow(downstream_cost.detach() - baseline, 2.0).sum()

    if use_baseline:
        if downstream_cost.shape != baseline.shape:
            raise ValueError(
                "Expected baseline at site {} to be {} instead got {}".format(
                    node, downstream_cost.shape, baseline.shape
                )
            )

    return use_baseline, baseline_loss, baseline


def _compute_downstream_costs(model_trace, guide_trace, non_reparam_nodes):  #
    # recursively compute downstream cost nodes for all sample sites in model and guide
    # (even though ultimately just need for non-reparameterizable sample sites)
    # 1. downstream costs used for rao-blackwellization
    # 2. model observe sites (as well as terms that arise from the model and guide having different
    # dependency structures) are taken care of via 'children_in_model' below
    topo_sort_guide_nodes = guide_trace.topological_sort(reverse=True)
    topo_sort_guide_nodes = [
        x for x in topo_sort_guide_nodes if guide_trace.nodes[x]["type"] == "sample"
    ]
    ordered_guide_nodes_dict = {n: i for i, n in enumerate(topo_sort_guide_nodes)}

    downstream_guide_cost_nodes = {}
    downstream_costs = {}
    stacks = get_plate_stacks(model_trace)

    for node in topo_sort_guide_nodes:
        downstream_costs[node] = MultiFrameTensor(
            (
                stacks[node],
                model_trace.nodes[node]["log_prob"]
                - guide_trace.nodes[node]["log_prob"],
            )
        )
        nodes_included_in_sum = set([node])
        downstream_guide_cost_nodes[node] = set([node])
        # make more efficient by ordering children appropriately (higher children first)
        children = [
            (k, -ordered_guide_nodes_dict[k]) for k in guide_trace.successors(node)
        ]
        sorted_children = sorted(children, key=itemgetter(1))
        for child, _ in sorted_children:
            child_cost_nodes = downstream_guide_cost_nodes[child]
            downstream_guide_cost_nodes[node].update(child_cost_nodes)
            if nodes_included_in_sum.isdisjoint(child_cost_nodes):  # avoid duplicates
                downstream_costs[node].add(*downstream_costs[child].items())
                # XXX nodes_included_in_sum logic could be more fine-grained, possibly leading
                # to speed-ups in case there are many duplicates
                nodes_included_in_sum.update(child_cost_nodes)
        missing_downstream_costs = (
            downstream_guide_cost_nodes[node] - nodes_included_in_sum
        )
        # include terms we missed because we had to avoid duplicates
        for missing_node in missing_downstream_costs:
            downstream_costs[node].add(
                (
                    stacks[missing_node],
                    model_trace.nodes[missing_node]["log_prob"]
                    - guide_trace.nodes[missing_node]["log_prob"],
                )
            )

    # finish assembling complete downstream costs
    # (the above computation may be missing terms from model)
    for site in non_reparam_nodes:
        children_in_model = set()
        for node in downstream_guide_cost_nodes[site]:
            children_in_model.update(model_trace.successors(node))
        # remove terms accounted for above
        children_in_model.difference_update(downstream_guide_cost_nodes[site])
        for child in children_in_model:
            assert model_trace.nodes[child]["type"] == "sample"
            downstream_costs[site].add(
                (stacks[child], model_trace.nodes[child]["log_prob"])
            )
            downstream_guide_cost_nodes[site].update([child])

    for k in non_reparam_nodes:
        downstream_costs[k] = downstream_costs[k].sum_to(
            guide_trace.nodes[k]["cond_indep_stack"]
        )

    return downstream_costs, downstream_guide_cost_nodes


def _compute_elbo(model_trace, guide_trace):
    # In ref [1], section 3.2, the part of the surrogate loss computed here is
    # \sum{cost}, which in this case is the ELBO. Instead of using the ELBO,
    # this implementation uses a surrogate ELBO which modifies some entropy
    # terms depending on the parameterization. This reduces the variance of the
    # gradient under some conditions.

    elbo = 0.0
    surrogate_elbo = 0.0
    baseline_loss = 0.0
    # mapping from non-reparameterizable sample sites to cost terms influenced by each of them
    downstream_costs = defaultdict(lambda: MultiFrameTensor())

    # Bring log p(x, z|...) terms into both the ELBO and the surrogate
    for name, site in model_trace.nodes.items():
        if site["type"] == "sample":
            elbo += site["log_prob_sum"]
            surrogate_elbo += site["log_prob_sum"]
            # add the log_prob to each non-reparam sample site upstream
            for key in get_provenance(site["log_prob_sum"]):
                downstream_costs[key].add((site["cond_indep_stack"], site["log_prob"]))

    # Bring log q(z|...) terms into the ELBO, and effective terms into the
    # surrogate. Depending on the parameterization of a site, its log q(z|...)
    # cost term may not contribute (in expectation) to the gradient. To reduce
    # the variance under some conditions, the default entropy terms from
    # site[`score_parts`] are used.
    for name, site in guide_trace.nodes.items():
        if site["type"] == "sample":
            elbo -= site["log_prob_sum"]
            entropy_term = site["score_parts"].entropy_term
            # For fully reparameterized terms, this entropy_term is log q(z|...)
            # For fully non-reparameterized terms, it is zero
            if not is_identically_zero(entropy_term):
                surrogate_elbo -= entropy_term.sum()
            # add the -log_prob to each non-reparam sample site upstream
            for key in get_provenance(site["log_prob_sum"]):
                downstream_costs[key].add((site["cond_indep_stack"], -site["log_prob"]))

    # construct all the reinforce-like terms.
    # we include only downstream costs to reduce variance
    # optionally include baselines to further reduce variance
    for node, downstream_cost in downstream_costs.items():
        guide_site = guide_trace.nodes[node]
        downstream_cost = downstream_cost.sum_to(guide_site["cond_indep_stack"])
        score_function = guide_site["score_parts"].score_function

        use_baseline, baseline_loss_term, baseline = _construct_baseline(
            node, guide_site, downstream_cost
        )

        if use_baseline:
            downstream_cost = downstream_cost - baseline
            baseline_loss = baseline_loss + baseline_loss_term

        surrogate_elbo += (score_function * downstream_cost.detach()).sum()

    surrogate_loss = -surrogate_elbo + baseline_loss
    return detach_provenance(elbo), detach_provenance(surrogate_loss)


class TrackNonReparam(Messenger):
    """
    Track non-reparameterizable sample sites.

    **References:**

    1. *Nonstandard Interpretations of Probabilistic Programs for Efficient Inference*,
        David Wingate, Noah Goodman, Andreas Stuhlmüller, Jeffrey Siskind

    **Example:**

    .. doctest::

       >>> import torch
       >>> import pyro
       >>> import pyro.distributions as dist
       >>> from pyro.infer.tracegraph_elbo import TrackNonReparam
       >>> from pyro.ops.provenance import get_provenance
       >>> from pyro.poutine import trace

       >>> def model():
       ...     probs_a = torch.tensor([0.3, 0.7])
       ...     probs_b = torch.tensor([[0.1, 0.9], [0.8, 0.2]])
       ...     probs_c = torch.tensor([[0.5, 0.5], [0.6, 0.4]])
       ...     a = pyro.sample("a", dist.Categorical(probs_a))
       ...     b = pyro.sample("b", dist.Categorical(probs_b[a]))
       ...     pyro.sample("c", dist.Categorical(probs_c[b]), obs=torch.tensor(0))

       >>> with TrackNonReparam():
       ...     model_tr = trace(model).get_trace()
       >>> model_tr.compute_log_prob()

       >>> print(get_provenance(model_tr.nodes["a"]["log_prob"]))  # doctest: +SKIP
       frozenset({'a'})
       >>> print(get_provenance(model_tr.nodes["b"]["log_prob"]))  # doctest: +SKIP
       frozenset({'b', 'a'})
       >>> print(get_provenance(model_tr.nodes["c"]["log_prob"]))  # doctest: +SKIP
       frozenset({'b', 'a'})
    """

    def _pyro_post_sample(self, msg):
        if (
            msg["type"] == "sample"
            and not isinstance(msg["fn"], _Subsample)
            and not msg["is_observed"]
            and not getattr(msg["fn"], "has_rsample", False)
        ):
            provenance = frozenset({msg["name"]})
            msg["value"] = track_provenance(msg["value"], provenance)


class TraceGraph_ELBO(ELBO):
    """
    A TraceGraph implementation of ELBO-based SVI. The gradient estimator
    is constructed along the lines of reference [1] specialized to the case
    of the ELBO. It supports arbitrary dependency structure for the model
    and guide as well as baselines for non-reparameterizable random variables.
    Fine-grained conditional dependency information as recorded in the
    :class:`~pyro.poutine.trace.Trace` is used to reduce the variance of the gradient estimator.
    In particular provenance tracking [3] is used to find the ``cost`` terms
    that depend on each non-reparameterizable sample site.

    References

    [1] `Gradient Estimation Using Stochastic Computation Graphs`,
        John Schulman, Nicolas Heess, Theophane Weber, Pieter Abbeel

    [2] `Neural Variational Inference and Learning in Belief Networks`
        Andriy Mnih, Karol Gregor

    [3] `Nonstandard Interpretations of Probabilistic Programs for Efficient Inference`,
        David Wingate, Noah Goodman, Andreas Stuhlmüller, Jeffrey Siskind
    """

    def _get_trace(self, model, guide, args, kwargs):
        """
        Returns a single trace from the guide, and the model that is run
        against it.
        """
        with TrackNonReparam():
            model_trace, guide_trace = get_importance_trace(
                "dense", self.max_plate_nesting, model, guide, args, kwargs
            )
        if is_validation_enabled():
            check_if_enumerated(guide_trace)
        return model_trace, guide_trace

    def loss(self, model, guide, *args, **kwargs):
        """
        :returns: returns an estimate of the ELBO
        :rtype: float

        Evaluates the ELBO with an estimator that uses num_particles many samples/particles.
        """
        elbo = 0.0
        for model_trace, guide_trace in self._get_traces(model, guide, args, kwargs):
            elbo_particle = torch_item(model_trace.log_prob_sum()) - torch_item(
                guide_trace.log_prob_sum()
            )
            elbo += elbo_particle / float(self.num_particles)

        loss = -elbo
        warn_if_nan(loss, "loss")
        return loss

    def loss_and_grads(self, model, guide, *args, **kwargs):
        """
        :returns: returns an estimate of the ELBO
        :rtype: float

        Computes the ELBO as well as the surrogate ELBO that is used to form the gradient estimator.
        Performs backward on the latter. Num_particle many samples are used to form the estimators.
        If baselines are present, a baseline loss is also constructed and differentiated.
        """
        elbo, surrogate_loss = self._loss_and_surrogate_loss(model, guide, args, kwargs)

        torch_backward(surrogate_loss, retain_graph=self.retain_graph)

        elbo = torch_item(elbo)
        loss = -elbo
        warn_if_nan(loss, "loss")
        return loss

    def _loss_and_surrogate_loss(self, model, guide, args, kwargs):
        loss = 0.0
        surrogate_loss = 0.0

        for model_trace, guide_trace in self._get_traces(model, guide, args, kwargs):
            lp, slp = self._loss_and_surrogate_loss_particle(model_trace, guide_trace)
            loss += lp
            surrogate_loss += slp

        loss /= self.num_particles
        surrogate_loss /= self.num_particles

        return loss, surrogate_loss

    def _loss_and_surrogate_loss_particle(self, model_trace, guide_trace):
        elbo, surrogate_loss = _compute_elbo(model_trace, guide_trace)

        return elbo, surrogate_loss


class JitTraceGraph_ELBO(TraceGraph_ELBO):
    """
    Like :class:`TraceGraph_ELBO` but uses :func:`torch.jit.trace` to
    compile :meth:`loss_and_grads`.

    This works only for a limited set of models:

    -   Models must have static structure.
    -   Models must not depend on any global data (except the param store).
    -   All model inputs that are tensors must be passed in via ``*args``.
    -   All model inputs that are *not* tensors must be passed in via
        ``**kwargs``, and compilation will be triggered once per unique
        ``**kwargs``.
    """

    def loss_and_grads(self, model, guide, *args, **kwargs):
        kwargs["_pyro_model_id"] = id(model)
        kwargs["_pyro_guide_id"] = id(guide)
        if getattr(self, "_jit_loss_and_surrogate_loss", None) is None:
            # build a closure for loss_and_surrogate_loss
            weakself = weakref.ref(self)

            @pyro.ops.jit.trace(
                ignore_warnings=self.ignore_jit_warnings, jit_options=self.jit_options
            )
            def jit_loss_and_surrogate_loss(*args, **kwargs):
                kwargs.pop("_pyro_model_id")
                kwargs.pop("_pyro_guide_id")
                self = weakself()
                return self._loss_and_surrogate_loss(model, guide, args, kwargs)

            self._jit_loss_and_surrogate_loss = jit_loss_and_surrogate_loss

        elbo, surrogate_loss = self._jit_loss_and_surrogate_loss(*args, **kwargs)

        surrogate_loss.backward(
            retain_graph=self.retain_graph
        )  # triggers jit compilation

        loss = -elbo.item()
        warn_if_nan(loss, "loss")
        return loss



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

from typing import Callable, Dict, Optional

import torch

import pyro
import pyro.poutine as poutine
from pyro.poutine.trace_struct import Trace
from pyro.poutine.util import site_is_subsample


class Resampler:
    """Resampler for interactive tuning of generative models, typically
    when preforming prior predictive checks as an early step of Bayesian
    workflow.

    This is intended as a computational cache to speed up the interactive
    tuning of the parameters of prior distributions based on samples from a
    downstream simulation. The idea is that the simulation can be expensive,
    but that when one slightly tweaks parameters of the parameter distribution
    then one can reuse most of the previous samples via importance resampling.

    :param callable guide: A pyro model that takes no arguments. The guide
        should be diffuse, covering more space than the subsequent ``model``
        passed to :meth:`sample`. Must be vectorizable via ``pyro.plate``.
    :param callable simulator: An optional larger pyro model with a superset of
        the guide's latent variables. Must be vectorizable via ``pyro.plate``.
    :param int num_guide_samples: Number of inital samples to draw from the
        guide. This should be much larger than the ``num_samples`` requested in
        subsequent calls to :meth:`sample`.
    :param int max_plate_nesting: The maximum plate nesting in the model.
        If absent this will be guessed by running the guide.
    """

    def __init__(
        self,
        guide: Callable,
        simulator: Optional[Callable] = None,
        *,
        num_guide_samples: int,
        max_plate_nesting: Optional[int] = None,
    ):
        super().__init__()
        if max_plate_nesting is None:
            max_plate_nesting = _guess_max_plate_nesting(
                guide if simulator is None else simulator
            )
        self._particle_dim = -1 - max_plate_nesting
        self._gumbels: Optional[torch.Tensor] = None

        # Draw samples from the initial guide.
        with pyro.plate("particles", num_guide_samples, dim=self._particle_dim):
            trace = poutine.trace(guide).get_trace()
            self._old_logp = _log_prob_sum(trace, num_guide_samples)

            if simulator:
                # Draw extended samples from the simulator.
                trace = poutine.trace(poutine.replay(simulator, trace)).get_trace()
        self._samples = {
            name: site["value"]
            for name, site in trace.nodes.items()
            if site["type"] == "sample" and not site_is_subsample(site)
        }

    @torch.no_grad()
    def sample(
        self, model: Callable, num_samples: int, stable: bool = True
    ) -> Dict[str, torch.Tensor]:
        """Draws a set of at most ``num_samples`` many model samples,
        optionally extended by the ``simulator``.

        Internally this importance resamples the samples generated by the
        ``guide`` in ``.__init__()``, and does not rerun the ``guide`` or
        ``simulator``. If the original guide samples poorly cover the model
        distribution, samples will show low diversity.

        :param callable model: A model with the same latent variables as the
            original ``guide``. Must be vectorizable via ``pyro.plate``.
        :param int num_samples: The number of samples to draw.
        :param bool stable: Whether to use piecewise-constant multinomial
            sampling. Set to True for visualization, False for Monte Carlo
            integration. Defaults to True.
        :returns: A dictionary of stacked samples.
        :rtype: Dict[str, torch.Tensor]
        """
        num_guide_samples = len(self._old_logp)
        with pyro.plate("particles", num_guide_samples, dim=self._particle_dim):
            trace = poutine.trace(poutine.condition(model, self._samples)).get_trace()
        new_logp = _log_prob_sum(trace, num_guide_samples)
        logits = new_logp - self._old_logp
        i = self._categorical_sample(logits, num_samples, stable)
        samples = {k: v[i] for k, v in self._samples.items()}
        return samples

    def _categorical_sample(
        self, logits: torch.Tensor, num_samples: int, stable: bool
    ) -> torch.Tensor:
        if not stable:
            return torch.multinomial(logits.exp(), num_samples, replacement=True)

        # Implement stable categorical sampling via the Gumbel-max trick.
        if self._gumbels is None or len(self._gumbels) < num_samples:
            # gumbel ~ -log(-log(uniform(0,1)))
            tiny = torch.finfo(logits.dtype).tiny
            self._gumbels = logits.new_empty(num_samples, len(logits)).uniform_()
            self._gumbels.clamp_(min=tiny).log_().neg_().clamp_(min=tiny).log_().neg_()
        return self._gumbels[:num_samples].add(logits).max(-1).indices


def _log_prob_sum(trace: Trace, batch_size: int) -> torch.Tensor:
    """Computes vectorized log_prob_sum batched over the leftmost dimension."""
    trace.compute_log_prob()
    result = 0.0
    for site in trace.nodes.values():
        if site["type"] == "sample":
            logp = site["log_prob"]
            assert logp.shape[:1] == (batch_size,)
            result += logp.reshape(batch_size, -1).sum(-1)
    return result


def _guess_max_plate_nesting(model: callable) -> int:
    with torch.no_grad(), poutine.block(), poutine.mask(mask=False):
        trace = poutine.trace(model).get_trace()
    plate_nesting = {0}.union(
        -f.dim
        for site in trace.nodes.values()
        for f in site.get("cond_indep_stack", [])
        if f.vectorized
    )
    return max(plate_nesting)



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import sys
from typing import TYPE_CHECKING, Callable, Generic, Literal, Optional, TypeVar

from typing_extensions import ParamSpec, Self

from pyro.poutine.messenger import Messenger
from pyro.poutine.trace_struct import Trace
from pyro.poutine.util import site_is_subsample

if TYPE_CHECKING:
    from pyro.poutine.runtime import Message

_P = ParamSpec("_P")
_T = TypeVar("_T")


def identify_dense_edges(trace: Trace) -> None:
    """
    Modifies a trace in-place by adding all edges based on the
    `cond_indep_stack` information stored at each site.
    """
    for name, node in trace.nodes.items():
        if site_is_subsample(node):
            continue
        if node["type"] == "sample":
            for past_name, past_node in trace.nodes.items():
                if site_is_subsample(past_node):
                    continue
                if past_node["type"] == "sample":
                    if past_name == name:
                        break
                    past_node_independent = False
                    for query, target in zip(
                        node["cond_indep_stack"], past_node["cond_indep_stack"]
                    ):
                        if (
                            query.name == target.name
                            and query.counter != target.counter
                        ):
                            past_node_independent = True
                            break
                    if not past_node_independent:
                        trace.add_edge(past_name, name)


class TraceMessenger(Messenger):
    """
    Return a handler that records the inputs and outputs of primitive calls
    and their dependencies.

    Consider the following Pyro program:

        >>> def model(x):
        ...     s = pyro.param("s", torch.tensor(0.5))
        ...     z = pyro.sample("z", dist.Normal(x, s))
        ...     return z ** 2

    We can record its execution using ``trace``
    and use the resulting data structure to compute the log-joint probability
    of all of the sample sites in the execution or extract all parameters.

        >>> trace = pyro.poutine.trace(model).get_trace(0.0)
        >>> logp = trace.log_prob_sum()
        >>> params = [trace.nodes[name]["value"].unconstrained() for name in trace.param_nodes]

    :param fn: a stochastic function (callable containing Pyro primitive calls)
    :param graph_type: string that specifies the kind of graph to construct
    :param param_only: if true, only records params and not samples
    :returns: stochastic function decorated with a :class:`~pyro.poutine.trace_messenger.TraceMessenger`
    """

    def __init__(
        self,
        graph_type: Optional[Literal["flat", "dense"]] = None,
        param_only: Optional[bool] = None,
    ) -> None:
        """
        :param string graph_type: string that specifies the type of graph
            to construct (currently only "flat" or "dense" supported)
        :param param_only: boolean that specifies whether to record sample sites
        """
        super().__init__()
        if graph_type is None:
            graph_type = "flat"
        if param_only is None:
            param_only = False
        assert graph_type in ("flat", "dense")
        self.graph_type = graph_type
        self.param_only = param_only
        self.trace = Trace(graph_type=self.graph_type)

    def __enter__(self) -> Self:
        self.trace = Trace(graph_type=self.graph_type)
        return super().__enter__()

    def __exit__(self, *args, **kwargs) -> None:
        """
        Adds appropriate edges based on cond_indep_stack information
        upon exiting the context.
        """
        if self.param_only:
            for node in list(self.trace.nodes.values()):
                if node["type"] != "param":
                    assert node["name"] is not None
                    self.trace.remove_node(node["name"])
        if self.graph_type == "dense":
            identify_dense_edges(self.trace)
        return super().__exit__(*args, **kwargs)

    def __call__(self, fn: Callable[_P, _T]) -> Callable[_P, _T]:
        """
        TODO docs
        """
        return TraceHandler(self, fn)

    def get_trace(self) -> Trace:
        """
        :returns: data structure
        :rtype: pyro.poutine.Trace

        Helper method for a very common use case.
        Returns a shallow copy of ``self.trace``.
        """
        return self.trace.copy()

    def _reset(self) -> None:
        tr = Trace(graph_type=self.graph_type)
        if "_INPUT" in self.trace.nodes:
            tr.add_node(
                "_INPUT",
                name="_INPUT",
                type="input",
                args=self.trace.nodes["_INPUT"]["args"],
                kwargs=self.trace.nodes["_INPUT"]["kwargs"],
            )
        self.trace = tr
        super()._reset()

    def _pyro_post_sample(self, msg: "Message") -> None:
        if self.param_only:
            return
        assert msg["name"] is not None
        assert msg["infer"] is not None
        if msg["infer"].get("_do_not_trace"):
            assert msg["infer"].get("is_auxiliary")
            assert not msg["is_observed"]
            return
        self.trace.add_node(msg["name"], **msg.copy())

    def _pyro_post_param(self, msg: "Message") -> None:
        assert msg["name"] is not None
        self.trace.add_node(msg["name"], **msg.copy())


class TraceHandler(Generic[_P, _T]):
    """
    Execution trace poutine.

    A TraceHandler records the input and output to every Pyro primitive
    and stores them as a site in a Trace().
    This should, in theory, be sufficient information for every inference algorithm
    (along with the implicit computational graph in the Variables?)

    We can also use this for visualization.
    """

    def __init__(self, msngr: TraceMessenger, fn: Callable[_P, _T]) -> None:
        self.fn = fn
        self.msngr = msngr

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _T:
        """
        Runs the stochastic function stored in this poutine,
        with additional side effects.

        Resets self.trace to an empty trace,
        installs itself on the global execution stack,
        runs self.fn with the given arguments,
        uninstalls itself from the global execution stack,
        stores the arguments and return value of the function in special sites,
        and returns self.fn's return value
        """
        with self.msngr:
            self.msngr.trace.add_node(
                "_INPUT", name="_INPUT", type="args", args=args, kwargs=kwargs
            )
            try:
                ret = self.fn(*args, **kwargs)
            except (ValueError, RuntimeError) as e:
                exc_type, exc_value, traceback = sys.exc_info()
                shapes = self.msngr.trace.format_shapes()
                assert exc_type is not None
                exc = exc_type("{}\n{}".format(exc_value, shapes))
                exc = exc.with_traceback(traceback)
                raise exc from e
            self.msngr.trace.add_node(
                "_RETURN", name="_RETURN", type="return", value=ret
            )
        return ret

    @property
    def trace(self) -> Trace:
        return self.msngr.trace

    def get_trace(self, *args, **kwargs) -> Trace:
        """
        :returns: data structure
        :rtype: pyro.poutine.Trace

        Helper method for a very common use case.
        Calls this poutine and returns its trace instead of the function's return value.
        """
        self(*args, **kwargs)
        return self.msngr.get_trace()



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import math
import warnings

import torch

import pyro
from pyro import poutine
from pyro.contrib.oed.search import Search
from pyro.contrib.util import lexpand
from pyro.infer import SVI, EmpiricalMarginal, Importance
from pyro.infer.autoguide.utils import mean_field_entropy
from pyro.util import torch_isinf, torch_isnan

__all__ = [
    "laplace_eig",
    "vi_eig",
    "nmc_eig",
    "donsker_varadhan_eig",
    "posterior_eig",
    "marginal_eig",
    "lfire_eig",
    "vnmc_eig",
]


def laplace_eig(
    model,
    design,
    observation_labels,
    target_labels,
    guide,
    loss,
    optim,
    num_steps,
    final_num_samples,
    y_dist=None,
    eig=True,
    **prior_entropy_kwargs
):
    """
    Estimates the expected information gain (EIG) by making repeated Laplace approximations to the posterior.

    :param function model: Pyro stochastic function taking `design` as only argument.
    :param torch.Tensor design: Tensor of possible designs.
    :param list observation_labels: labels of sample sites to be regarded as observables.
    :param list target_labels: labels of sample sites to be regarded as latent variables of interest, i.e. the sites
        that we wish to gain information about.
    :param function guide: Pyro stochastic function corresponding to `model`.
    :param loss: a Pyro loss such as `pyro.infer.Trace_ELBO().differentiable_loss`.
    :param optim: optimizer for the loss
    :param int num_steps: Number of gradient steps to take per sampled pseudo-observation.
    :param int final_num_samples: Number of `y` samples (pseudo-observations) to take.
    :param y_dist: Distribution to sample `y` from- if `None` we use the Bayesian marginal distribution.
    :param bool eig: Whether to compute the EIG or the average posterior entropy (APE). The EIG is given by
        `EIG = prior entropy - APE`. If `True`, the prior entropy will be estimated analytically,
        or by Monte Carlo as appropriate for the `model`. If `False` the APE is returned.
    :param dict prior_entropy_kwargs: parameters for estimating the prior entropy: `num_prior_samples` indicating the
        number of samples for a MC estimate of prior entropy, and `mean_field` indicating if an analytic form for
        a mean-field prior should be tried.
    :return: EIG estimate, optionally includes full optimization history
    :rtype: torch.Tensor
    """

    if isinstance(observation_labels, str):
        observation_labels = [observation_labels]
    if target_labels is not None and isinstance(target_labels, str):
        target_labels = [target_labels]

    ape = _laplace_vi_ape(
        model,
        design,
        observation_labels,
        target_labels,
        guide,
        loss,
        optim,
        num_steps,
        final_num_samples,
        y_dist=y_dist,
    )
    return _eig_from_ape(model, design, target_labels, ape, eig, prior_entropy_kwargs)


def _eig_from_ape(model, design, target_labels, ape, eig, prior_entropy_kwargs):
    mean_field = prior_entropy_kwargs.get("mean_field", True)
    if eig:
        if mean_field:
            try:
                prior_entropy = mean_field_entropy(
                    model, [design], whitelist=target_labels
                )
            except NotImplemented:
                prior_entropy = monte_carlo_entropy(
                    model, design, target_labels, **prior_entropy_kwargs
                )
        else:
            prior_entropy = monte_carlo_entropy(
                model, design, target_labels, **prior_entropy_kwargs
            )
        return prior_entropy - ape
    else:
        return ape


def _laplace_vi_ape(
    model,
    design,
    observation_labels,
    target_labels,
    guide,
    loss,
    optim,
    num_steps,
    final_num_samples,
    y_dist=None,
):
    def posterior_entropy(y_dist, design):
        # Important that y_dist is sampled *within* the function
        y = pyro.sample("conditioning_y", y_dist)
        y_dict = {label: y[i, ...] for i, label in enumerate(observation_labels)}
        conditioned_model = pyro.condition(model, data=y_dict)
        # Here just using SVI to run the MAP optimization
        guide.train()
        svi = SVI(conditioned_model, guide=guide, loss=loss, optim=optim)
        with poutine.block():
            for _ in range(num_steps):
                svi.step(design)
        # Recover the entropy
        with poutine.block():
            final_loss = loss(conditioned_model, guide, design)
            guide.finalize(final_loss, target_labels)
            entropy = mean_field_entropy(guide, [design], whitelist=target_labels)
        return entropy

    if y_dist is None:
        y_dist = EmpiricalMarginal(
            Importance(model, num_samples=final_num_samples).run(design),
            sites=observation_labels,
        )

    # Calculate the expected posterior entropy under this distn of y
    loss_dist = EmpiricalMarginal(Search(posterior_entropy).run(y_dist, design))
    ape = loss_dist.mean

    return ape


# Deprecated
def vi_eig(
    model,
    design,
    observation_labels,
    target_labels,
    vi_parameters,
    is_parameters,
    y_dist=None,
    eig=True,
    **prior_entropy_kwargs
):
    """.. deprecated:: 0.4.1
        Use `posterior_eig` instead.

    Estimates the expected information gain (EIG) using variational inference (VI).

    The APE is defined as

        :math:`APE(d)=E_{Y\\sim p(y|\\theta, d)}[H(p(\\theta|Y, d))]`

    where :math:`H[p(x)]` is the `differential entropy
    <https://en.wikipedia.org/wiki/Differential_entropy>`_.
    The APE is related to expected information gain (EIG) by the equation

        :math:`EIG(d)=H[p(\\theta)]-APE(d)`

    in particular, minimising the APE is equivalent to maximising EIG.

    :param function model: A pyro model accepting `design` as only argument.
    :param torch.Tensor design: Tensor representation of design
    :param list observation_labels: A subset of the sample sites
        present in `model`. These sites are regarded as future observations
        and other sites are regarded as latent variables over which a
        posterior is to be inferred.
    :param list target_labels: A subset of the sample sites over which the posterior
        entropy is to be measured.
    :param dict vi_parameters: Variational inference parameters which should include:
        `optim`: an instance of :class:`pyro.Optim`, `guide`: a guide function
        compatible with `model`, `num_steps`: the number of VI steps to make,
        and `loss`: the loss function to use for VI
    :param dict is_parameters: Importance sampling parameters for the
        marginal distribution of :math:`Y`. May include `num_samples`: the number
        of samples to draw from the marginal.
    :param pyro.distributions.Distribution y_dist: (optional) the distribution
        assumed for the response variable :math:`Y`
    :param bool eig: Whether to compute the EIG or the average posterior entropy (APE). The EIG is given by
        `EIG = prior entropy - APE`. If `True`, the prior entropy will be estimated analytically,
        or by Monte Carlo as appropriate for the `model`. If `False` the APE is returned.
    :param dict prior_entropy_kwargs: parameters for estimating the prior entropy: `num_prior_samples` indicating the
        number of samples for a MC estimate of prior entropy, and `mean_field` indicating if an analytic form for
        a mean-field prior should be tried.
    :return: EIG estimate, optionally includes full optimization history
    :rtype: torch.Tensor

    """

    warnings.warn(
        "`vi_eig` is deprecated in favour of the amortized version: `posterior_eig`.",
        DeprecationWarning,
    )

    if isinstance(observation_labels, str):
        observation_labels = [observation_labels]
    if target_labels is not None and isinstance(target_labels, str):
        target_labels = [target_labels]

    ape = _vi_ape(
        model,
        design,
        observation_labels,
        target_labels,
        vi_parameters,
        is_parameters,
        y_dist=y_dist,
    )
    return _eig_from_ape(model, design, target_labels, ape, eig, prior_entropy_kwargs)


def _vi_ape(
    model,
    design,
    observation_labels,
    target_labels,
    vi_parameters,
    is_parameters,
    y_dist=None,
):
    svi_num_steps = vi_parameters.pop("num_steps")

    def posterior_entropy(y_dist, design):
        # Important that y_dist is sampled *within* the function
        y = pyro.sample("conditioning_y", y_dist)
        y_dict = {label: y[i, ...] for i, label in enumerate(observation_labels)}
        conditioned_model = pyro.condition(model, data=y_dict)
        svi = SVI(conditioned_model, **vi_parameters)
        with poutine.block():
            for _ in range(svi_num_steps):
                svi.step(design)
        # Recover the entropy
        with poutine.block():
            guide = vi_parameters["guide"]
            entropy = mean_field_entropy(guide, [design], whitelist=target_labels)
        return entropy

    if y_dist is None:
        y_dist = EmpiricalMarginal(
            Importance(model, **is_parameters).run(design), sites=observation_labels
        )

    # Calculate the expected posterior entropy under this distn of y
    loss_dist = EmpiricalMarginal(Search(posterior_entropy).run(y_dist, design))
    loss = loss_dist.mean

    return loss


def nmc_eig(
    model,
    design,
    observation_labels,
    target_labels=None,
    N=100,
    M=10,
    M_prime=None,
    independent_priors=False,
):
    """Nested Monte Carlo estimate of the expected information
    gain (EIG). The estimate is, when there are not any random effects,

    .. math::

        \\frac{1}{N}\\sum_{n=1}^N \\log p(y_n | \\theta_n, d) -
        \\frac{1}{N}\\sum_{n=1}^N \\log \\left(\\frac{1}{M}\\sum_{m=1}^M p(y_n | \\theta_m, d)\\right)

    where :math:`\\theta_n, y_n \\sim p(\\theta, y | d)` and :math:`\\theta_m \\sim p(\\theta)`.
    The estimate in the presence of random effects is

    .. math::

        \\frac{1}{N}\\sum_{n=1}^N  \\log \\left(\\frac{1}{M'}\\sum_{m=1}^{M'}
        p(y_n | \\theta_n, \\widetilde{\\theta}_{nm}, d)\\right)-
        \\frac{1}{N}\\sum_{n=1}^N \\log \\left(\\frac{1}{M}\\sum_{m=1}^{M}
        p(y_n | \\theta_m, \\widetilde{\\theta}_{m}, d)\\right)

    where :math:`\\widetilde{\\theta}` are the random effects with
    :math:`\\widetilde{\\theta}_{nm} \\sim p(\\widetilde{\\theta}|\\theta=\\theta_n)` and
    :math:`\\theta_m,\\widetilde{\\theta}_m \\sim p(\\theta,\\widetilde{\\theta})`.
    The latter form is used when `M_prime != None`.

    :param function model: A pyro model accepting `design` as only argument.
    :param torch.Tensor design: Tensor representation of design
    :param list observation_labels: A subset of the sample sites
        present in `model`. These sites are regarded as future observations
        and other sites are regarded as latent variables over which a
        posterior is to be inferred.
    :param list target_labels: A subset of the sample sites over which the posterior
        entropy is to be measured.
    :param int N: Number of outer expectation samples.
    :param int M: Number of inner expectation samples for `p(y|d)`.
    :param int M_prime: Number of samples for `p(y | theta, d)` if required.
    :param bool independent_priors: Only used when `M_prime` is not `None`. Indicates whether the prior distributions
        for the target variables and the nuisance variables are independent. In this case, it is not necessary to
        sample the targets conditional on the nuisance variables.
    :return: EIG estimate, optionally includes full optimization history
    :rtype: torch.Tensor
    """

    if isinstance(observation_labels, str):  # list of strings instead of strings
        observation_labels = [observation_labels]
    if isinstance(target_labels, str):
        target_labels = [target_labels]

    # Take N samples of the model
    expanded_design = lexpand(design, N)  # N copies of the model
    trace = poutine.trace(model).get_trace(expanded_design)
    trace.compute_log_prob()

    if M_prime is not None:
        y_dict = {
            l: lexpand(trace.nodes[l]["value"], M_prime) for l in observation_labels
        }
        theta_dict = {
            l: lexpand(trace.nodes[l]["value"], M_prime) for l in target_labels
        }
        theta_dict.update(y_dict)
        # Resample M values of u and compute conditional probabilities
        # WARNING: currently the use of condition does not actually sample
        # the conditional distribution!
        # We need to use some importance weighting
        conditional_model = pyro.condition(model, data=theta_dict)
        if independent_priors:
            reexpanded_design = lexpand(design, M_prime, 1)
        else:
            # Not acceptable to use (M_prime, 1) here - other variables may occur after
            # theta, so need to be sampled conditional upon it
            reexpanded_design = lexpand(design, M_prime, N)
        retrace = poutine.trace(conditional_model).get_trace(reexpanded_design)
        retrace.compute_log_prob()
        conditional_lp = sum(
            retrace.nodes[l]["log_prob"] for l in observation_labels
        ).logsumexp(0) - math.log(M_prime)
    else:
        # This assumes that y are independent conditional on theta
        # Furthermore assume that there are no other variables besides theta
        conditional_lp = sum(trace.nodes[l]["log_prob"] for l in observation_labels)

    y_dict = {l: lexpand(trace.nodes[l]["value"], M) for l in observation_labels}
    # Resample M values of theta and compute conditional probabilities
    conditional_model = pyro.condition(model, data=y_dict)
    # Using (M, 1) instead of (M, N) - acceptable to re-use thetas between ys because
    # theta comes before y in graphical model
    reexpanded_design = lexpand(design, M, 1)  # sample M theta
    retrace = poutine.trace(conditional_model).get_trace(reexpanded_design)
    retrace.compute_log_prob()
    marginal_lp = sum(
        retrace.nodes[l]["log_prob"] for l in observation_labels
    ).logsumexp(0) - math.log(M)

    terms = conditional_lp - marginal_lp
    nonnan = (~torch.isnan(terms)).sum(0).type_as(terms)
    terms[torch.isnan(terms)] = 0.0
    return terms.sum(0) / nonnan


def donsker_varadhan_eig(
    model,
    design,
    observation_labels,
    target_labels,
    num_samples,
    num_steps,
    T,
    optim,
    return_history=False,
    final_design=None,
    final_num_samples=None,
):
    """
    Donsker-Varadhan estimate of the expected information gain (EIG).

    The Donsker-Varadhan representation of EIG is

    .. math::

        \\sup_T E_{p(y, \\theta | d)}[T(y, \\theta)] - \\log E_{p(y|d)p(\\theta)}[\\exp(T(\\bar{y}, \\bar{\\theta}))]

    where :math:`T` is any (measurable) function.

    This methods optimises the loss function over a pre-specified class of
    functions `T`.

    :param function model: A pyro model accepting `design` as only argument.
    :param torch.Tensor design: Tensor representation of design
    :param list observation_labels: A subset of the sample sites
        present in `model`. These sites are regarded as future observations
        and other sites are regarded as latent variables over which a
        posterior is to be inferred.
    :param list target_labels: A subset of the sample sites over which the posterior
        entropy is to be measured.
    :param int num_samples: Number of samples per iteration.
    :param int num_steps: Number of optimization steps.
    :param function or torch.nn.Module T: optimisable function `T` for use in the
        Donsker-Varadhan loss function.
    :param pyro.optim.Optim optim: Optimiser to use.
    :param bool return_history: If `True`, also returns a tensor giving the loss function
        at each step of the optimization.
    :param torch.Tensor final_design: The final design tensor to evaluate at. If `None`, uses
        `design`.
    :param int final_num_samples: The number of samples to use at the final evaluation, If `None,
        uses `num_samples`.
    :return: EIG estimate, optionally includes full optimization history
    :rtype: torch.Tensor or tuple
    """
    if isinstance(observation_labels, str):
        observation_labels = [observation_labels]
    if isinstance(target_labels, str):
        target_labels = [target_labels]
    loss = _donsker_varadhan_loss(model, T, observation_labels, target_labels)
    return opt_eig_ape_loss(
        design,
        loss,
        num_samples,
        num_steps,
        optim,
        return_history,
        final_design,
        final_num_samples,
    )


def posterior_eig(
    model,
    design,
    observation_labels,
    target_labels,
    num_samples,
    num_steps,
    guide,
    optim,
    return_history=False,
    final_design=None,
    final_num_samples=None,
    eig=True,
    prior_entropy_kwargs={},
    *args,
    **kwargs
):
    """
    Posterior estimate of expected information gain (EIG) computed from the average posterior entropy (APE)
    using :math:`EIG(d) = H[p(\\theta)] - APE(d)`. See [1] for full details.

    The posterior representation of APE is

        :math:`\\sup_{q}\\ E_{p(y, \\theta | d)}[\\log q(\\theta | y, d)]`

    where :math:`q` is any distribution on :math:`\\theta`.

    This method optimises the loss over a given `guide` family representing :math:`q`.

    [1] Foster, Adam, et al. "Variational Bayesian Optimal Experimental Design." arXiv preprint arXiv:1903.05480 (2019).

    :param function model: A pyro model accepting `design` as only argument.
    :param torch.Tensor design: Tensor representation of design
    :param list observation_labels: A subset of the sample sites
        present in `model`. These sites are regarded as future observations
        and other sites are regarded as latent variables over which a
        posterior is to be inferred.
    :param list target_labels: A subset of the sample sites over which the posterior
        entropy is to be measured.
    :param int num_samples: Number of samples per iteration.
    :param int num_steps: Number of optimization steps.
    :param function guide: guide family for use in the (implicit) posterior estimation.
        The parameters of `guide` are optimised to maximise the posterior
        objective.
    :param pyro.optim.Optim optim: Optimiser to use.
    :param bool return_history: If `True`, also returns a tensor giving the loss function
        at each step of the optimization.
    :param torch.Tensor final_design: The final design tensor to evaluate at. If `None`, uses
        `design`.
    :param int final_num_samples: The number of samples to use at the final evaluation, If `None,
        uses `num_samples`.
    :param bool eig: Whether to compute the EIG or the average posterior entropy (APE). The EIG is given by
        `EIG = prior entropy - APE`. If `True`, the prior entropy will be estimated analytically,
        or by Monte Carlo as appropriate for the `model`. If `False` the APE is returned.
    :param dict prior_entropy_kwargs: parameters for estimating the prior entropy: `num_prior_samples` indicating the
        number of samples for a MC estimate of prior entropy, and `mean_field` indicating if an analytic form for
        a mean-field prior should be tried.
    :return: EIG estimate, optionally includes full optimization history
    :rtype: torch.Tensor or tuple
    """
    if isinstance(observation_labels, str):
        observation_labels = [observation_labels]
    if isinstance(target_labels, str):
        target_labels = [target_labels]

    ape = _posterior_ape(
        model,
        design,
        observation_labels,
        target_labels,
        num_samples,
        num_steps,
        guide,
        optim,
        return_history=return_history,
        final_design=final_design,
        final_num_samples=final_num_samples,
        *args,
        **kwargs
    )
    return _eig_from_ape(model, design, target_labels, ape, eig, prior_entropy_kwargs)


def _posterior_ape(
    model,
    design,
    observation_labels,
    target_labels,
    num_samples,
    num_steps,
    guide,
    optim,
    return_history=False,
    final_design=None,
    final_num_samples=None,
    *args,
    **kwargs
):
    loss = _posterior_loss(
        model, guide, observation_labels, target_labels, *args, **kwargs
    )
    return opt_eig_ape_loss(
        design,
        loss,
        num_samples,
        num_steps,
        optim,
        return_history,
        final_design,
        final_num_samples,
    )


def marginal_eig(
    model,
    design,
    observation_labels,
    target_labels,
    num_samples,
    num_steps,
    guide,
    optim,
    return_history=False,
    final_design=None,
    final_num_samples=None,
):
    """Estimate EIG by estimating the marginal entropy :math:`p(y|d)`. See [1] for full details.

    The marginal representation of EIG is

        :math:`\\inf_{q}\\ E_{p(y, \\theta | d)}\\left[\\log \\frac{p(y | \\theta, d)}{q(y | d)} \\right]`

    where :math:`q` is any distribution on :math:`y`. A variational family for :math:`q` is specified in the `guide`.

    .. warning :: This method does **not** estimate the correct quantity in the presence of random effects.

    [1] Foster, Adam, et al. "Variational Bayesian Optimal Experimental Design." arXiv preprint arXiv:1903.05480 (2019).

    :param function model: A pyro model accepting `design` as only argument.
    :param torch.Tensor design: Tensor representation of design
    :param list observation_labels: A subset of the sample sites
        present in `model`. These sites are regarded as future observations
        and other sites are regarded as latent variables over which a
        posterior is to be inferred.
    :param list target_labels: A subset of the sample sites over which the posterior
        entropy is to be measured.
    :param int num_samples: Number of samples per iteration.
    :param int num_steps: Number of optimization steps.
    :param function guide: guide family for use in the marginal estimation.
        The parameters of `guide` are optimised to maximise the log-likelihood objective.
    :param pyro.optim.Optim optim: Optimiser to use.
    :param bool return_history: If `True`, also returns a tensor giving the loss function
        at each step of the optimization.
    :param torch.Tensor final_design: The final design tensor to evaluate at. If `None`, uses
        `design`.
    :param int final_num_samples: The number of samples to use at the final evaluation, If `None,
        uses `num_samples`.
    :return: EIG estimate, optionally includes full optimization history
    :rtype: torch.Tensor or tuple
    """

    if isinstance(observation_labels, str):
        observation_labels = [observation_labels]
    if isinstance(target_labels, str):
        target_labels = [target_labels]
    loss = _marginal_loss(model, guide, observation_labels, target_labels)
    return opt_eig_ape_loss(
        design,
        loss,
        num_samples,
        num_steps,
        optim,
        return_history,
        final_design,
        final_num_samples,
    )


def marginal_likelihood_eig(
    model,
    design,
    observation_labels,
    target_labels,
    num_samples,
    num_steps,
    marginal_guide,
    cond_guide,
    optim,
    return_history=False,
    final_design=None,
    final_num_samples=None,
):
    """Estimates EIG by estimating the marginal entropy, that of :math:`p(y|d)`,
    *and* the conditional entropy, of :math:`p(y|\\theta, d)`, both via Gibbs' Inequality. See [1] for full details.

    [1] Foster, Adam, et al. "Variational Bayesian Optimal Experimental Design." arXiv preprint arXiv:1903.05480 (2019).

    :param function model: A pyro model accepting `design` as only argument.
    :param torch.Tensor design: Tensor representation of design
    :param list observation_labels: A subset of the sample sites
        present in `model`. These sites are regarded as future observations
        and other sites are regarded as latent variables over which a
        posterior is to be inferred.
    :param list target_labels: A subset of the sample sites over which the posterior
        entropy is to be measured.
    :param int num_samples: Number of samples per iteration.
    :param int num_steps: Number of optimization steps.
    :param function marginal_guide: guide family for use in the marginal estimation.
        The parameters of `guide` are optimised to maximise the log-likelihood objective.
    :param function cond_guide: guide family for use in the likelihood (conditional) estimation.
        The parameters of `guide` are optimised to maximise the log-likelihood objective.
    :param pyro.optim.Optim optim: Optimiser to use.
    :param bool return_history: If `True`, also returns a tensor giving the loss function
        at each step of the optimization.
    :param torch.Tensor final_design: The final design tensor to evaluate at. If `None`, uses
        `design`.
    :param int final_num_samples: The number of samples to use at the final evaluation, If `None,
        uses `num_samples`.
    :return: EIG estimate, optionally includes full optimization history
    :rtype: torch.Tensor or tuple
    """

    if isinstance(observation_labels, str):
        observation_labels = [observation_labels]
    if isinstance(target_labels, str):
        target_labels = [target_labels]
    loss = _marginal_likelihood_loss(
        model, marginal_guide, cond_guide, observation_labels, target_labels
    )
    return opt_eig_ape_loss(
        design,
        loss,
        num_samples,
        num_steps,
        optim,
        return_history,
        final_design,
        final_num_samples,
    )


def lfire_eig(
    model,
    design,
    observation_labels,
    target_labels,
    num_y_samples,
    num_theta_samples,
    num_steps,
    classifier,
    optim,
    return_history=False,
    final_design=None,
    final_num_samples=None,
):
    """Estimates the EIG using the method of Likelihood-Free Inference by Ratio Estimation (LFIRE) as in [1].
    LFIRE is run separately for several samples of :math:`\\theta`.

    [1] Kleinegesse, Steven, and Michael Gutmann. "Efficient Bayesian Experimental Design for Implicit Models."
    arXiv preprint arXiv:1810.09912 (2018).

    :param function model: A pyro model accepting `design` as only argument.
    :param torch.Tensor design: Tensor representation of design
    :param list observation_labels: A subset of the sample sites
        present in `model`. These sites are regarded as future observations
        and other sites are regarded as latent variables over which a
        posterior is to be inferred.
    :param list target_labels: A subset of the sample sites over which the posterior
        entropy is to be measured.
    :param int num_y_samples: Number of samples to take in :math:`y` for each :math:`\\theta`.
    :param: int num_theta_samples: Number of initial samples in :math:`\\theta` to take. The likelihood ratio
                                   is estimated by LFIRE for each sample.
    :param int num_steps: Number of optimization steps.
    :param function classifier: a Pytorch or Pyro classifier used to distinguish between samples of :math:`y` under
                                :math:`p(y|d)` and samples under :math:`p(y|\\theta,d)` for some :math:`\\theta`.
    :param pyro.optim.Optim optim: Optimiser to use.
    :param bool return_history: If `True`, also returns a tensor giving the loss function
        at each step of the optimization.
    :param torch.Tensor final_design: The final design tensor to evaluate at. If `None`, uses
        `design`.
    :param int final_num_samples: The number of samples to use at the final evaluation, If `None,
        uses `num_samples`.
    :return: EIG estimate, optionally includes full optimization history
    :rtype: torch.Tensor or tuple
    """
    if isinstance(observation_labels, str):
        observation_labels = [observation_labels]
    if isinstance(target_labels, str):
        target_labels = [target_labels]

    # Take N samples of the model
    expanded_design = lexpand(design, num_theta_samples)
    trace = poutine.trace(model).get_trace(expanded_design)

    theta_dict = {l: trace.nodes[l]["value"] for l in target_labels}
    cond_model = pyro.condition(model, data=theta_dict)

    loss = _lfire_loss(model, cond_model, classifier, observation_labels, target_labels)
    out = opt_eig_ape_loss(
        expanded_design,
        loss,
        num_y_samples,
        num_steps,
        optim,
        return_history,
        final_design,
        final_num_samples,
    )
    if return_history:
        return out[0], out[1].sum(0) / num_theta_samples
    else:
        return out.sum(0) / num_theta_samples


def vnmc_eig(
    model,
    design,
    observation_labels,
    target_labels,
    num_samples,
    num_steps,
    guide,
    optim,
    return_history=False,
    final_design=None,
    final_num_samples=None,
):
    """Estimates the EIG using Variational Nested Monte Carlo (VNMC). The VNMC estimate [1] is

    .. math::

        \\frac{1}{N}\\sum_{n=1}^N \\left[ \\log p(y_n | \\theta_n, d) -
         \\log \\left(\\frac{1}{M}\\sum_{m=1}^M \\frac{p(\\theta_{mn})p(y_n | \\theta_{mn}, d)}
         {q(\\theta_{mn} | y_n)} \\right) \\right]

    where :math:`q(\\theta | y)` is the learned variational posterior approximation and
    :math:`\\theta_n, y_n \\sim p(\\theta, y | d)`, :math:`\\theta_{mn} \\sim q(\\theta|y=y_n)`.

    As :math:`N \\to \\infty` this is an upper bound on EIG. We minimise this upper bound by stochastic gradient
    descent.

    .. warning :: This method cannot be used in the presence of random effects.

    [1] Foster, Adam, et al. "Variational Bayesian Optimal Experimental Design." arXiv preprint arXiv:1903.05480 (2019).

    :param function model: A pyro model accepting `design` as only argument.
    :param torch.Tensor design: Tensor representation of design
    :param list observation_labels: A subset of the sample sites
        present in `model`. These sites are regarded as future observations
        and other sites are regarded as latent variables over which a
        posterior is to be inferred.
    :param list target_labels: A subset of the sample sites over which the posterior
        entropy is to be measured.
    :param tuple num_samples: Number of (:math:`N, M`) samples per iteration.
    :param int num_steps: Number of optimization steps.
    :param function guide: guide family for use in the posterior estimation.
        The parameters of `guide` are optimised to minimise the VNMC upper bound.
    :param pyro.optim.Optim optim: Optimiser to use.
    :param bool return_history: If `True`, also returns a tensor giving the loss function
        at each step of the optimization.
    :param torch.Tensor final_design: The final design tensor to evaluate at. If `None`, uses
        `design`.
    :param tuple final_num_samples: The number of (:math:`N, M`) samples to use at the final evaluation, If `None,
        uses `num_samples`.
    :return: EIG estimate, optionally includes full optimization history
    :rtype: torch.Tensor or tuple
    """
    if isinstance(observation_labels, str):
        observation_labels = [observation_labels]
    if isinstance(target_labels, str):
        target_labels = [target_labels]
    loss = _vnmc_eig_loss(model, guide, observation_labels, target_labels)
    return opt_eig_ape_loss(
        design,
        loss,
        num_samples,
        num_steps,
        optim,
        return_history,
        final_design,
        final_num_samples,
    )


def opt_eig_ape_loss(
    design,
    loss_fn,
    num_samples,
    num_steps,
    optim,
    return_history=False,
    final_design=None,
    final_num_samples=None,
):
    if final_design is None:
        final_design = design
    if final_num_samples is None:
        final_num_samples = num_samples

    params = None
    history = []
    for step in range(num_steps):
        if params is not None:
            pyro.infer.util.zero_grads(params)
        with poutine.trace(param_only=True) as param_capture:
            agg_loss, loss = loss_fn(design, num_samples, evaluation=return_history)
        params = set(
            site["value"].unconstrained() for site in param_capture.trace.nodes.values()
        )
        if torch.isnan(agg_loss):
            raise ArithmeticError("Encountered NaN loss in opt_eig_ape_loss")
        agg_loss.backward(retain_graph=True)
        if return_history:
            history.append(loss)
        optim(params)
        try:
            optim.step()
        except AttributeError:
            pass

    _, loss = loss_fn(final_design, final_num_samples, evaluation=True)
    if return_history:
        return torch.stack(history), loss
    else:
        return loss


def monte_carlo_entropy(model, design, target_labels, num_prior_samples=1000):
    """Computes a Monte Carlo estimate of the entropy of `model` assuming that each of sites in `target_labels` is
    independent and the entropy is to be computed for that subset of sites only.
    """

    if isinstance(target_labels, str):
        target_labels = [target_labels]

    expanded_design = lexpand(design, num_prior_samples)
    trace = pyro.poutine.trace(model).get_trace(expanded_design)
    trace.compute_log_prob()
    lp = sum(trace.nodes[l]["log_prob"] for l in target_labels)
    return -lp.sum(0) / num_prior_samples


def _donsker_varadhan_loss(model, T, observation_labels, target_labels):
    """DV loss: to evaluate directly use `donsker_varadhan_eig` setting `num_steps=0`."""

    ewma_log = EwmaLog(alpha=0.90)

    def loss_fn(design, num_particles, **kwargs):
        try:
            pyro.module("T", T)
        except AssertionError:
            pass

        expanded_design = lexpand(design, num_particles)

        # Unshuffled data
        unshuffled_trace = poutine.trace(model).get_trace(expanded_design)
        y_dict = {l: unshuffled_trace.nodes[l]["value"] for l in observation_labels}

        # Shuffled data
        # Not actually shuffling, resimulate for safety
        conditional_model = pyro.condition(model, data=y_dict)
        shuffled_trace = poutine.trace(conditional_model).get_trace(expanded_design)

        T_joint = T(
            expanded_design, unshuffled_trace, observation_labels, target_labels
        )
        T_independent = T(
            expanded_design, shuffled_trace, observation_labels, target_labels
        )

        joint_expectation = T_joint.sum(0) / num_particles

        A = T_independent - math.log(num_particles)
        s, _ = torch.max(A, dim=0)
        independent_expectation = s + ewma_log((A - s).exp().sum(dim=0), s)

        loss = joint_expectation - independent_expectation
        # Switch sign, sum over batch dimensions for scalar loss
        agg_loss = -loss.sum()
        return agg_loss, loss

    return loss_fn


def _posterior_loss(
    model, guide, observation_labels, target_labels, analytic_entropy=False
):
    """Posterior loss: to evaluate directly use `posterior_eig` setting `num_steps=0`, `eig=False`."""

    def loss_fn(design, num_particles, evaluation=False, **kwargs):
        expanded_design = lexpand(design, num_particles)

        # Sample from p(y, theta | d)
        trace = poutine.trace(model).get_trace(expanded_design)
        y_dict = {l: trace.nodes[l]["value"] for l in observation_labels}
        theta_dict = {l: trace.nodes[l]["value"] for l in target_labels}

        # Run through q(theta | y, d)
        conditional_guide = pyro.condition(guide, data=theta_dict)
        cond_trace = poutine.trace(conditional_guide).get_trace(
            y_dict, expanded_design, observation_labels, target_labels
        )
        cond_trace.compute_log_prob()
        if evaluation and analytic_entropy:
            loss = (
                mean_field_entropy(
                    guide,
                    [y_dict, expanded_design, observation_labels, target_labels],
                    whitelist=target_labels,
                ).sum(0)
                / num_particles
            )
            agg_loss = loss.sum()
        else:
            terms = -sum(cond_trace.nodes[l]["log_prob"] for l in target_labels)
            agg_loss, loss = _safe_mean_terms(terms)

        return agg_loss, loss

    return loss_fn


def _marginal_loss(model, guide, observation_labels, target_labels):
    """Marginal loss: to evaluate directly use `marginal_eig` setting `num_steps=0`."""

    def loss_fn(design, num_particles, evaluation=False, **kwargs):
        expanded_design = lexpand(design, num_particles)

        # Sample from p(y | d)
        trace = poutine.trace(model).get_trace(expanded_design)
        y_dict = {l: trace.nodes[l]["value"] for l in observation_labels}

        # Run through q(y | d)
        conditional_guide = pyro.condition(guide, data=y_dict)
        cond_trace = poutine.trace(conditional_guide).get_trace(
            expanded_design, observation_labels, target_labels
        )
        cond_trace.compute_log_prob()

        terms = -sum(cond_trace.nodes[l]["log_prob"] for l in observation_labels)

        # At eval time, add p(y | theta, d) terms
        if evaluation:
            trace.compute_log_prob()
            terms += sum(trace.nodes[l]["log_prob"] for l in observation_labels)

        return _safe_mean_terms(terms)

    return loss_fn


def _marginal_likelihood_loss(
    model, marginal_guide, likelihood_guide, observation_labels, target_labels
):
    """Marginal_likelihood loss: to evaluate directly use `marginal_likelihood_eig` setting `num_steps=0`."""

    def loss_fn(design, num_particles, evaluation=False, **kwargs):
        expanded_design = lexpand(design, num_particles)

        # Sample from p(y | d)
        trace = poutine.trace(model).get_trace(expanded_design)
        y_dict = {l: trace.nodes[l]["value"] for l in observation_labels}
        theta_dict = {l: trace.nodes[l]["value"] for l in target_labels}

        # Run through q(y | d)
        qyd = pyro.condition(marginal_guide, data=y_dict)
        marginal_trace = poutine.trace(qyd).get_trace(
            expanded_design, observation_labels, target_labels
        )
        marginal_trace.compute_log_prob()

        # Run through q(y | theta, d)
        qythetad = pyro.condition(likelihood_guide, data=y_dict)
        cond_trace = poutine.trace(qythetad).get_trace(
            theta_dict, expanded_design, observation_labels, target_labels
        )
        cond_trace.compute_log_prob()
        terms = -sum(marginal_trace.nodes[l]["log_prob"] for l in observation_labels)

        # At evaluation time, use the right estimator, q(y | theta, d) - q(y | d)
        # At training time, use -q(y | theta, d) - q(y | d) so gradients go the same way
        if evaluation:
            terms += sum(cond_trace.nodes[l]["log_prob"] for l in observation_labels)
        else:
            terms -= sum(cond_trace.nodes[l]["log_prob"] for l in observation_labels)

        return _safe_mean_terms(terms)

    return loss_fn


def _lfire_loss(
    model_marginal, model_conditional, h, observation_labels, target_labels
):
    """LFIRE loss: to evaluate directly use `lfire_eig` setting `num_steps=0`."""

    def loss_fn(design, num_particles, evaluation=False, **kwargs):
        try:
            pyro.module("h", h)
        except AssertionError:
            pass

        expanded_design = lexpand(design, num_particles)
        model_conditional_trace = poutine.trace(model_conditional).get_trace(
            expanded_design
        )

        if not evaluation:
            model_marginal_trace = poutine.trace(model_marginal).get_trace(
                expanded_design
            )

            h_joint = h(
                expanded_design,
                model_conditional_trace,
                observation_labels,
                target_labels,
            )
            h_independent = h(
                expanded_design, model_marginal_trace, observation_labels, target_labels
            )

            terms = torch.nn.functional.softplus(
                -h_joint
            ) + torch.nn.functional.softplus(h_independent)
            return _safe_mean_terms(terms)

        else:
            h_joint = h(
                expanded_design,
                model_conditional_trace,
                observation_labels,
                target_labels,
            )
            return _safe_mean_terms(h_joint)

    return loss_fn


def _vnmc_eig_loss(model, guide, observation_labels, target_labels):
    """VNMC loss: to evaluate directly use `vnmc_eig` setting `num_steps=0`."""

    def loss_fn(design, num_particles, evaluation=False, **kwargs):
        N, M = num_particles
        expanded_design = lexpand(design, N)

        # Sample from p(y, theta | d)
        trace = poutine.trace(model).get_trace(expanded_design)
        y_dict = {l: lexpand(trace.nodes[l]["value"], M) for l in observation_labels}

        # Sample M times from q(theta | y, d) for each y
        reexpanded_design = lexpand(expanded_design, M)
        conditional_guide = pyro.condition(guide, data=y_dict)
        guide_trace = poutine.trace(conditional_guide).get_trace(
            y_dict, reexpanded_design, observation_labels, target_labels
        )
        theta_y_dict = {l: guide_trace.nodes[l]["value"] for l in target_labels}
        theta_y_dict.update(y_dict)
        guide_trace.compute_log_prob()

        # Re-run that through the model to compute the joint
        modelp = pyro.condition(model, data=theta_y_dict)
        model_trace = poutine.trace(modelp).get_trace(reexpanded_design)
        model_trace.compute_log_prob()

        terms = -sum(guide_trace.nodes[l]["log_prob"] for l in target_labels)
        terms += sum(model_trace.nodes[l]["log_prob"] for l in target_labels)
        terms += sum(model_trace.nodes[l]["log_prob"] for l in observation_labels)
        terms = -terms.logsumexp(0) + math.log(M)

        # At eval time, add p(y | theta, d) terms
        if evaluation:
            trace.compute_log_prob()
            terms += sum(trace.nodes[l]["log_prob"] for l in observation_labels)

        return _safe_mean_terms(terms)

    return loss_fn


def _safe_mean_terms(terms):
    mask = torch.isnan(terms) | (terms == float("-inf")) | (terms == float("inf"))
    if terms.dtype is torch.float32:
        nonnan = (~mask).sum(0).float()
    elif terms.dtype is torch.float64:
        nonnan = (~mask).sum(0).double()
    terms[mask] = 0.0
    loss = terms.sum(0) / nonnan
    agg_loss = loss.sum()
    return agg_loss, loss


def xexpx(a):
    """Computes `a*exp(a)`.

    This function makes the outputs more stable when the inputs of this function converge to :math:`-\\infty`.

    :param torch.Tensor a:
    :return: Equivalent of `a*torch.exp(a)`.
    """
    mask = a == float("-inf")
    y = a * torch.exp(a)
    y[mask] = 0.0
    return y


class _EwmaLogFn(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, ewma):
        ctx.save_for_backward(ewma)
        return input.log()

    @staticmethod
    def backward(ctx, grad_output):
        (ewma,) = ctx.saved_tensors
        return grad_output / ewma, None


_ewma_log_fn = _EwmaLogFn.apply


class EwmaLog:
    """Logarithm function with exponentially weighted moving average
    for gradients.

    For input `inputs` this function return :code:`inputs.log()`. However, it
    computes the gradient as

        :math:`\\frac{\\sum_{t=0}^{T-1} \\alpha^t}{\\sum_{t=0}^{T-1} \\alpha^t x_{T-t}}`

    where :math:`x_t` are historical input values passed to this function,
    :math:`x_T` being the most recently seen value.

    This gradient may help with numerical stability when the sequence of
    inputs to the function form a convergent sequence.
    """

    def __init__(self, alpha):
        self.alpha = alpha
        self.ewma = 0.0
        self.n = 0
        self.s = 0.0

    def __call__(self, inputs, s, dim=0, keepdim=False):
        """Updates the moving average, and returns :code:`inputs.log()`."""
        self.n += 1
        if torch_isnan(self.ewma) or torch_isinf(self.ewma):
            ewma = inputs
        else:
            ewma = inputs * (1.0 - self.alpha) / (1 - self.alpha**self.n) + torch.exp(
                self.s - s
            ) * self.ewma * (self.alpha - self.alpha**self.n) / (
                1 - self.alpha**self.n
            )
        self.ewma = ewma.detach()
        self.s = s.detach()
        return _ewma_log_fn(inputs, ewma)



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Dict, List, Optional

import torch
from typing_extensions import Self

from pyro.distributions.torch import Categorical
from pyro.distributions.torch_distribution import TorchDistributionMixin
from pyro.ops.indexing import Vindex
from pyro.poutine.messenger import Messenger
from pyro.poutine.runtime import _ENUM_ALLOCATOR, Message
from pyro.util import ignore_jit_warnings


def _tmc_mixture_sample(msg: Message) -> torch.Tensor:
    assert isinstance(msg["fn"], TorchDistributionMixin)
    assert msg["infer"] is not None
    dist, num_samples = msg["fn"], msg["infer"]["num_samples"]

    # find batch dims that aren't plate dims
    batch_shape = [1] * len(dist.batch_shape)
    for f in msg["cond_indep_stack"]:
        if f.vectorized:
            assert f.dim is not None
            batch_shape[f.dim] = f.size if f.size > 0 else dist.batch_shape[f.dim]
    batch_shape_tuple = tuple(batch_shape)

    # sample a batch
    sample_shape = (num_samples,)
    fat_sample = dist(
        sample_shape=torch.Size(sample_shape)
    )  # TODO thin before sampling
    assert fat_sample.shape == sample_shape + dist.batch_shape + dist.event_shape
    assert any(d > 1 for d in fat_sample.shape)

    target_shape = (num_samples,) + batch_shape_tuple + dist.event_shape

    # if this site has any possible ancestors, sample ancestor indices uniformly
    thin_sample = fat_sample
    if thin_sample.shape != target_shape:
        index = [Ellipsis] + [slice(None)] * (len(thin_sample.shape) - 1)
        squashed_dims = []
        for squashed_dim, squashed_size in zip(
            range(1, len(thin_sample.shape)), thin_sample.shape[1:]
        ):
            if squashed_size > 1 and (
                target_shape[squashed_dim] == 1 or squashed_dim == 0
            ):
                # uniformly sample one ancestor per upstream particle population
                ancestor_dist = Categorical(
                    logits=torch.zeros((squashed_size,), device=thin_sample.device)
                )
                ancestor_index = ancestor_dist.sample(sample_shape=(num_samples,))
                index[squashed_dim] = ancestor_index
                squashed_dims.append(squashed_dim)

        thin_sample = Vindex(thin_sample)[tuple(index)]
        for squashed_dim in squashed_dims:
            thin_sample = thin_sample.unsqueeze(squashed_dim)

    assert thin_sample.shape == target_shape
    return thin_sample


def _tmc_diagonal_sample(msg: Message) -> torch.Tensor:
    assert isinstance(msg["fn"], TorchDistributionMixin)
    assert msg["infer"] is not None
    dist, num_samples = msg["fn"], msg["infer"]["num_samples"]

    # find batch dims that aren't plate dims
    batch_shape = [1] * len(dist.batch_shape)
    for f in msg["cond_indep_stack"]:
        if f.vectorized:
            assert f.dim is not None
            batch_shape[f.dim] = f.size if f.size > 0 else dist.batch_shape[f.dim]
    batch_shape_tuple = tuple(batch_shape)

    # sample a batch
    sample_shape = (num_samples,)
    fat_sample = dist(
        sample_shape=torch.Size(sample_shape)
    )  # TODO thin before sampling
    assert fat_sample.shape == sample_shape + dist.batch_shape + dist.event_shape
    assert any(d > 1 for d in fat_sample.shape)

    target_shape = (num_samples,) + batch_shape_tuple + dist.event_shape

    # if this site has any ancestors, choose ancestors from diagonal approximation
    thin_sample = fat_sample
    if thin_sample.shape != target_shape:
        index: List[Any] = [Ellipsis] + [slice(None)] * (len(thin_sample.shape) - 1)
        squashed_dims = []
        for squashed_dim, squashed_size in zip(
            range(1, len(thin_sample.shape)), thin_sample.shape[1:]
        ):
            if squashed_size > 1 and (
                target_shape[squashed_dim] == 1 or squashed_dim == 0
            ):
                # diagonal approximation: identify particle indices across populations
                ancestor_index = torch.arange(squashed_size, device=thin_sample.device)
                index[squashed_dim] = ancestor_index
                squashed_dims.append(squashed_dim)

        thin_sample = Vindex(thin_sample)[tuple(index)]
        for squashed_dim in squashed_dims:
            thin_sample = thin_sample.unsqueeze(squashed_dim)

    assert thin_sample.shape == target_shape
    return thin_sample


def enumerate_site(msg: Message) -> torch.Tensor:
    assert isinstance(msg["fn"], TorchDistributionMixin)
    assert msg["infer"] is not None
    dist, num_samples = msg["fn"], msg["infer"].get("num_samples")
    if num_samples is None:
        # Enumerate over the support of the distribution.
        value = dist.enumerate_support(expand=msg["infer"].get("expand", False))
    elif num_samples > 1 and not msg["infer"].get("expand", False):
        tmc_strategy = msg["infer"].get("tmc", "diagonal")
        if tmc_strategy == "mixture":
            value = _tmc_mixture_sample(msg)
        elif tmc_strategy == "diagonal":
            value = _tmc_diagonal_sample(msg)
        else:
            raise ValueError("{} not a valid TMC strategy".format(tmc_strategy))
    elif num_samples > 1 and msg["infer"]["expand"]:
        # Monte Carlo sample the distribution.
        value = dist(sample_shape=torch.Size([num_samples]))
    assert value.dim() == 1 + len(dist.batch_shape) + len(dist.event_shape)
    return value


class EnumMessenger(Messenger):
    """
    Enumerates in parallel over discrete sample sites marked
    ``infer={"enumerate": "parallel"}``.

    :param int first_available_dim: The first tensor dimension (counting
        from the right) that is available for parallel enumeration. This
        dimension and all dimensions left may be used internally by Pyro.
        This should be a negative integer or None.
    """

    def __init__(self, first_available_dim: Optional[int] = None) -> None:
        assert (
            first_available_dim is None or first_available_dim < 0
        ), first_available_dim
        self.first_available_dim = first_available_dim
        super().__init__()

    def __enter__(self) -> Self:
        if self.first_available_dim is not None:
            _ENUM_ALLOCATOR.set_first_available_dim(self.first_available_dim)
        self._markov_depths: Dict[str, int] = (
            {}
        )  # site name -> depth (nonnegative integer)
        self._param_dims: Dict[str, Dict[int, int]] = (
            {}
        )  # site name -> (enum dim -> unique id)
        self._value_dims: Dict[str, Dict[int, int]] = (
            {}
        )  # site name -> (enum dim -> unique id)
        return super().__enter__()

    @ignore_jit_warnings()
    def _pyro_sample(self, msg: Message) -> None:
        """
        :param msg: current message at a trace site.
        :returns: a sample from the stochastic function at the site.
        """
        if msg["done"] or not isinstance(msg["fn"], TorchDistributionMixin):
            return

        assert isinstance(msg["name"], str)
        assert msg["infer"] is not None
        # Compute upstream dims in scope; these are unsafe to use for this site's target_dim.
        scope = msg["infer"].get("_markov_scope")  # site name -> markov depth
        param_dims = _ENUM_ALLOCATOR.dim_to_id.copy()  # enum dim -> unique id
        if scope is not None:
            for name, depth in scope.items():
                if (
                    self._markov_depths[name] == depth
                ):  # hide sites whose markov context has exited
                    param_dims.update(self._value_dims[name])
            self._markov_depths[msg["name"]] = msg["infer"]["_markov_depth"]
        self._param_dims[msg["name"]] = param_dims
        if msg["is_observed"] or msg["infer"].get("enumerate") != "parallel":
            return

        # Compute an enumerated value (at an arbitrary dim).
        value = enumerate_site(msg)
        actual_dim = -1 - len(msg["fn"].batch_shape)  # the leftmost dim of log_prob

        # Move actual_dim to a safe target_dim.
        target_dim, id_ = _ENUM_ALLOCATOR.allocate(
            None if scope is None else set(param_dims)
        )
        event_dim = msg["fn"].event_dim
        categorical_support = getattr(value, "_pyro_categorical_support", None)
        if categorical_support is not None:
            # Preserve categorical supports to speed up Categorical.log_prob().
            # See pyro/distributions/torch.py for details.
            assert target_dim < 0
            value = value.reshape(value.shape[:1] + (1,) * (-1 - target_dim))
            value._pyro_categorical_support = categorical_support  # type: ignore[attr-defined]
        elif actual_dim < target_dim:
            assert (
                value.size(target_dim - event_dim) == 1
            ), "pyro.markov dim conflict at dim {}".format(actual_dim)
            value = value.transpose(target_dim - event_dim, actual_dim - event_dim)
            while value.dim() and value.size(0) == 1:
                value = value.squeeze(0)
        elif target_dim < actual_dim:
            diff = actual_dim - target_dim
            value = value.reshape(value.shape[:1] + (1,) * diff + value.shape[1:])

        # Compute dims passed downstream through the value.
        value_dims = {
            dim: param_dims[dim]
            for dim in range(event_dim - value.dim(), 0)
            if value.size(dim - event_dim) > 1 and dim in param_dims
        }
        value_dims[target_dim] = id_

        msg["infer"]["_enumerate_dim"] = target_dim
        msg["infer"]["_dim_to_id"] = value_dims
        msg["value"] = value
        msg["done"] = True

    def _pyro_post_sample(self, msg: Message) -> None:
        # Save all dims exposed in this sample value.
        # Whereas all of site["_dim_to_id"] are needed to interpret a
        # site's log_prob tensor, only a filtered subset self._value_dims[msg["name"]]
        # are needed to interpret a site's value.
        if not isinstance(msg["fn"], TorchDistributionMixin):
            return
        value = msg["value"]
        if value is None:
            return

        assert isinstance(msg["name"], str)
        assert msg["infer"] is not None
        shape = value.data.shape[: value.dim() - msg["fn"].event_dim]
        dim_to_id = msg["infer"].setdefault("_dim_to_id", {})
        dim_to_id.update(self._param_dims.get(msg["name"], {}))
        with ignore_jit_warnings():
            self._value_dims[msg["name"]] = {
                dim: id_
                for dim, id_ in dim_to_id.items()
                if len(shape) >= -dim and shape[dim] > 1
            }



# Copyright (c) 2017-2019 Uber Technologies, Inc.
# SPDX-License-Identifier: Apache-2.0

import functools
import math
import numbers
import random
import sys
import timeit
import warnings
from collections import defaultdict
from contextlib import contextmanager
from itertools import zip_longest

import numpy as np
import torch

from pyro.poutine.util import site_is_subsample


def set_rng_seed(rng_seed):
    """
    Sets seeds of `torch` and `torch.cuda` (if available).

    :param int rng_seed: The seed value.
    """
    torch.manual_seed(rng_seed)
    random.seed(rng_seed)
    np.random.seed(rng_seed)


def get_rng_state():
    return {
        "torch": torch.get_rng_state(),
        "random": random.getstate(),
        "numpy": np.random.get_state(),
    }


def set_rng_state(state):
    torch.set_rng_state(state["torch"])
    random.setstate(state["random"])
    if "numpy" in state:
        import numpy as np

        np.random.set_state(state["numpy"])


def torch_isnan(x):
    """
    A convenient function to check if a Tensor contains any nan; also works with numbers
    """
    if isinstance(x, numbers.Number):
        return x != x
    return torch.isnan(x).any()


def torch_isinf(x):
    """
    A convenient function to check if a Tensor contains any +inf; also works with numbers
    """
    if isinstance(x, numbers.Number):
        return x == math.inf or x == -math.inf
    return (x == math.inf).any() or (x == -math.inf).any()


def warn_if_nan(value, msg="", *, filename=None, lineno=None):
    """
    A convenient function to warn if a Tensor or its grad contains any nan,
    also works with numbers.
    """
    if filename is None:
        try:
            frame = sys._getframe(1)
        except ValueError:
            filename = "sys"
            lineno = 1
        else:
            filename = frame.f_code.co_filename
            lineno = frame.f_lineno

    if torch.is_tensor(value) and value.requires_grad:
        value.register_hook(
            lambda x: warn_if_nan(
                x, "backward " + msg, filename=filename, lineno=lineno
            )
        )

    if torch_isnan(value):
        warnings.warn_explicit(
            "Encountered NaN{}".format(": " + msg if msg else "."),
            UserWarning,
            filename,
            lineno,
        )

    return value


def warn_if_inf(
    value, msg="", allow_posinf=False, allow_neginf=False, *, filename=None, lineno=None
):
    """
    A convenient function to warn if a Tensor or its grad contains any inf,
    also works with numbers.
    """
    if filename is None:
        try:
            frame = sys._getframe(1)
        except ValueError:
            filename = "sys"
            lineno = 1
        else:
            filename = frame.f_code.co_filename
            lineno = frame.f_lineno

    if torch.is_tensor(value) and value.requires_grad:
        value.register_hook(
            lambda x: warn_if_inf(
                x,
                "backward " + msg,
                allow_posinf,
                allow_neginf,
                filename=filename,
                lineno=lineno,
            )
        )

    if (not allow_posinf) and (
        value == math.inf
        if isinstance(value, numbers.Number)
        else (value == math.inf).any()
    ):
        warnings.warn_explicit(
            "Encountered +inf{}".format(": " + msg if msg else "."),
            UserWarning,
            filename,
            lineno,
        )
    if (not allow_neginf) and (
        value == -math.inf
        if isinstance(value, numbers.Number)
        else (value == -math.inf).any()
    ):
        warnings.warn_explicit(
            "Encountered -inf{}".format(": " + msg if msg else "."),
            UserWarning,
            filename,
            lineno,
        )

    return value


def save_visualization(trace, graph_output):
    """
    DEPRECATED Use :func:`pyro.infer.inspect.render_model()` instead.

    Take a trace generated by poutine.trace with `graph_type='dense'`
    and render the graph with the output saved to file.

    - non-reparameterized stochastic nodes are salmon
    - reparameterized stochastic nodes are half salmon, half grey
    - observation nodes are green

    :param pyro.poutine.Trace trace: a trace to be visualized
    :param graph_output: the graph will be saved to graph_output.pdf
    :type graph_output: str

    Example::

        trace = pyro.poutine.trace(model, graph_type="dense").get_trace()
        save_visualization(trace, 'output')
    """
    warnings.warn(
        "`save_visualization` function is deprecated and will be removed in "
        "a future version."
    )

    import graphviz

    g = graphviz.Digraph()

    for label, node in trace.nodes.items():
        if site_is_subsample(node):
            continue
        shape = "ellipse"
        if label in trace.stochastic_nodes and label not in trace.reparameterized_nodes:
            fillcolor = "salmon"
        elif label in trace.reparameterized_nodes:
            fillcolor = "lightgrey;.5:salmon"
        elif label in trace.observation_nodes:
            fillcolor = "darkolivegreen3"
        else:
            # only visualize RVs
            continue
        g.node(label, label=label, shape=shape, style="filled", fillcolor=fillcolor)

    for label1, label2 in trace.edges:
        if site_is_subsample(trace.nodes[label1]):
            continue
        if site_is_subsample(trace.nodes[label2]):
            continue
        g.edge(label1, label2)

    g.render(graph_output, view=False, cleanup=True)


def check_traces_match(trace1, trace2):
    """
    :param pyro.poutine.Trace trace1: Trace object of the model
    :param pyro.poutine.Trace trace2: Trace object of the guide
    :raises: RuntimeWarning, ValueError

    Checks that (1) there is a bijection between the samples in the two traces
    and (2) at each sample site two traces agree on sample shape.
    """
    # Check ordinary sample sites.
    vars1 = set(name for name, site in trace1.nodes.items() if site["type"] == "sample")
    vars2 = set(name for name, site in trace2.nodes.items() if site["type"] == "sample")
    if vars1 != vars2:
        warnings.warn("Model vars changed: {} vs {}".format(vars1, vars2))

    # Check shapes agree.
    for name in vars1:
        site1 = trace1.nodes[name]
        site2 = trace2.nodes[name]
        if hasattr(site1["fn"], "shape") and hasattr(site2["fn"], "shape"):
            shape1 = site1["fn"].shape(*site1["args"], **site1["kwargs"])
            shape2 = site2["fn"].shape(*site2["args"], **site2["kwargs"])
            if shape1 != shape2:
                raise ValueError(
                    "Site dims disagree at site '{}': {} vs {}".format(
                        name, shape1, shape2
                    )
                )


def check_model_guide_match(model_trace, guide_trace, max_plate_nesting=math.inf):
    """
    :param pyro.poutine.Trace model_trace: Trace object of the model
    :param pyro.poutine.Trace guide_trace: Trace object of the guide
    :raises: RuntimeWarning, ValueError

    Checks the following assumptions:
    1. Each sample site in the model also appears in the guide and is not
        marked auxiliary.
    2. Each sample site in the guide either appears in the model or is marked,
        auxiliary via ``infer={'is_auxiliary': True}``.
    3. Each :class:``~pyro.plate`` statement in the guide also appears in the
        model.
    4. At each sample site that appears in both the model and guide, the model
        and guide agree on sample shape.
    5. Model-side sequential enumeration is not implemented.
    """
    # Check ordinary sample sites.
    guide_vars = set(
        name
        for name, site in guide_trace.nodes.items()
        if site["type"] == "sample"
        if type(site["fn"]).__name__ != "_Subsample"
    )
    aux_vars = set(
        name
        for name, site in guide_trace.nodes.items()
        if site["type"] == "sample"
        if site["infer"].get("is_auxiliary")
    )
    model_vars = set(
        name
        for name, site in model_trace.nodes.items()
        if site["type"] == "sample" and not site["is_observed"]
        if type(site["fn"]).__name__ != "_Subsample"
    )
    enum_vars = set(
        name
        for name, site in model_trace.nodes.items()
        if site["type"] == "sample" and not site["is_observed"]
        if type(site["fn"]).__name__ != "_Subsample"
        if site["infer"].get("_enumerate_dim") is not None
        if name not in guide_vars
    )
    if aux_vars & model_vars:
        warnings.warn(
            "Found auxiliary vars in the model: {}".format(aux_vars & model_vars)
        )
    if not (guide_vars <= model_vars | aux_vars):
        warnings.warn(
            "Found non-auxiliary vars in guide but not model, "
            "consider marking these infer={{'is_auxiliary': True}}:\n{}".format(
                guide_vars - aux_vars - model_vars
            )
        )
    if not (model_vars <= guide_vars | enum_vars):
        bad_sites = model_vars - guide_vars - enum_vars
        for name in bad_sites:
            if model_trace.nodes[name]["infer"].get("enumerate") == "sequential":
                raise NotImplementedError(
                    f"At site {repr(name)}, "
                    "model-side sequential enumeration is not implemented. "
                    "Try parallel enumeration or guide-side enumeration."
                )
        warnings.warn(f"Found vars in model but not guide: {bad_sites}")

    # Check shapes agree.
    for name in model_vars & guide_vars:
        model_site = model_trace.nodes[name]
        guide_site = guide_trace.nodes[name]

        if hasattr(model_site["fn"], "event_dim") and hasattr(
            guide_site["fn"], "event_dim"
        ):
            if model_site["fn"].event_dim != guide_site["fn"].event_dim:
                raise ValueError(
                    "Model and guide event_dims disagree at site '{}': {} vs {}".format(
                        name, model_site["fn"].event_dim, guide_site["fn"].event_dim
                    )
                )

        if hasattr(model_site["fn"], "shape") and hasattr(guide_site["fn"], "shape"):
            model_shape = model_site["fn"].shape(
                *model_site["args"], **model_site["kwargs"]
            )
            guide_shape = guide_site["fn"].shape(
                *guide_site["args"], **guide_site["kwargs"]
            )
            if model_shape == guide_shape:
                continue

            # Allow broadcasting outside of max_plate_nesting.
            if len(model_shape) > max_plate_nesting:
                model_shape = model_shape[
                    len(model_shape) - max_plate_nesting - model_site["fn"].event_dim :
                ]
            if len(guide_shape) > max_plate_nesting:
                guide_shape = guide_shape[
                    len(guide_shape) - max_plate_nesting - guide_site["fn"].event_dim :
                ]
            if model_shape == guide_shape:
                continue
            for model_size, guide_size in zip_longest(
                reversed(model_shape), reversed(guide_shape), fillvalue=1
            ):
                if model_size != guide_size:
                    raise ValueError(
                        "Model and guide shapes disagree at site '{}': {} vs {}".format(
                            name, model_shape, guide_shape
                        )
                    )

    # Check subsample sites introduced by plate.
    model_vars = set(
        name
        for name, site in model_trace.nodes.items()
        if site["type"] == "sample" and not site["is_observed"]
        if type(site["fn"]).__name__ == "_Subsample"
    )
    guide_vars = set(
        name
        for name, site in guide_trace.nodes.items()
        if site["type"] == "sample"
        if type(site["fn"]).__name__ == "_Subsample"
    )
    if not (guide_vars <= model_vars):
        warnings.warn(
            "Found plate statements in guide but not model: {}".format(
                guide_vars - model_vars
            )
        )

    # Check factor statements in guide specify has_rsample.
    for name, site in guide_trace.nodes.items():
        if not site["type"] == "sample":
            continue
        if not site["infer"].get("is_auxiliary"):
            continue
        if type(site["fn"]).__name__ != "Unit":
            continue
        if "has_rsample" not in site["fn"].__dict__:
            raise ValueError(
                f'At guide site pyro.factor("{name}",...), '
                "missing specification of has_rsample. "
                "Please either set has_rsample=True if the factor statement arises "
                "from reparametrized sampling or has_rsample=False otherwise."
            )


def check_site_shape(site, max_plate_nesting):
    actual_shape = list(site["log_prob"].shape)

    # Compute expected shape.
    expected_shape = []
    for f in site["cond_indep_stack"]:
        if f.dim is not None:
            # Use the specified plate dimension, which counts from the right.
            assert f.dim < 0
            if len(expected_shape) < -f.dim:
                expected_shape = [None] * (
                    -f.dim - len(expected_shape)
                ) + expected_shape
            if expected_shape[f.dim] is not None:
                raise ValueError(
                    "\n  ".join(
                        [
                            'at site "{}" within plate("{}", dim={}), dim collision'.format(
                                site["name"], f.name, f.dim
                            ),
                            "Try setting dim arg in other plates.",
                        ]
                    )
                )
            expected_shape[f.dim] = f.size
    expected_shape = [-1 if e is None else e for e in expected_shape]

    # Check for plate stack overflow.
    if len(expected_shape) > max_plate_nesting:
        raise ValueError(
            "\n  ".join(
                [
                    'at site "{}", plate stack overflow'.format(site["name"]),
                    "Try increasing max_plate_nesting to at least {}".format(
                        len(expected_shape)
                    ),
                ]
            )
        )

    # Ignore dimensions left of max_plate_nesting.
    if max_plate_nesting < len(actual_shape):
        actual_shape = actual_shape[len(actual_shape) - max_plate_nesting :]

    # Check for incorrect plate placement on the right of max_plate_nesting.
    for actual_size, expected_size in zip_longest(
        reversed(actual_shape), reversed(expected_shape), fillvalue=1
    ):
        if expected_size != -1 and expected_size != actual_size:
            raise ValueError(
                "\n  ".join(
                    [
                        'at site "{}", invalid log_prob shape'.format(site["name"]),
                        "Expected {}, actual {}".format(expected_shape, actual_shape),
                        "Try one of the following fixes:",
                        "- enclose the batched tensor in a with pyro.plate(...): context",
                        "- .to_event(...) the distribution being sampled",
                        "- .permute() data dimensions",
                    ]
                )
            )

    # Check parallel dimensions on the left of max_plate_nesting.
    enum_dim = site["infer"].get("_enumerate_dim")
    if enum_dim is not None:
        if (
            len(site["fn"].batch_shape) >= -enum_dim
            and site["fn"].batch_shape[enum_dim] != 1
        ):
            raise ValueError(
                "\n  ".join(
                    [
                        'Enumeration dim conflict at site "{}"'.format(site["name"]),
                        "Try increasing pyro.markov history size",
                    ]
                )
            )


def _are_independent(counters1, counters2):
    for name, counter1 in counters1.items():
        if name in counters2:
            if counters2[name] != counter1:
                return True
    return False


def check_traceenum_requirements(model_trace, guide_trace):
    """
    Warn if user could easily rewrite the model or guide in a way that would
    clearly avoid invalid dependencies on enumerated variables.

    :class:`~pyro.infer.traceenum_elbo.TraceEnum_ELBO` enumerates over
    synchronized products rather than full cartesian products. Therefore models
    must ensure that no variable outside of an plate depends on an enumerated
    variable inside that plate. Since full dependency checking is impossible,
    this function aims to warn only in cases where models can be easily
    rewitten to be obviously correct.
    """
    enumerated_sites = set(
        name
        for name, site in guide_trace.nodes.items()
        if site["type"] == "sample" and site["infer"].get("enumerate")
    )
    for role, trace in [("model", model_trace), ("guide", guide_trace)]:
        plate_counters = {}  # for sequential plates only
        enumerated_contexts = defaultdict(set)
        for name, site in trace.nodes.items():
            if site["type"] != "sample":
                continue
            plate_counter = {
                f.name: f.counter for f in site["cond_indep_stack"] if not f.vectorized
            }
            context = frozenset(f for f in site["cond_indep_stack"] if f.vectorized)

            # Check that sites outside each independence context precede enumerated sites inside that context.
            for enumerated_context, names in enumerated_contexts.items():
                if not (context < enumerated_context):
                    continue
                names = sorted(
                    n
                    for n in names
                    if not _are_independent(plate_counter, plate_counters[n])
                )
                if not names:
                    continue
                diff = sorted(f.name for f in enumerated_context - context)
                warnings.warn(
                    "\n  ".join(
                        [
                            'at {} site "{}", possibly invalid dependency.'.format(
                                role, name
                            ),
                            'Expected site "{}" to precede sites "{}"'.format(
                                name, '", "'.join(sorted(names))
                            ),
                            'to avoid breaking independence of plates "{}"'.format(
                                '", "'.join(diff)
                            ),
                        ]
                    ),
                    RuntimeWarning,
                )

            plate_counters[name] = plate_counter
            if name in enumerated_sites:
                enumerated_contexts[context].add(name)


def check_if_enumerated(guide_trace):
    enumerated_sites = [
        name
        for name, site in guide_trace.nodes.items()
        if site["type"] == "sample" and site["infer"].get("enumerate")
    ]
    if enumerated_sites:
        warnings.warn(
            "\n".join(
                [
                    "Found sample sites configured for enumeration:"
                    ", ".join(enumerated_sites),
                    "If you want to enumerate sites, you need to use TraceEnum_ELBO instead.",
                ]
            )
        )


@contextmanager
def ignore_jit_warnings(filter=None):
    """
    Ignore JIT tracer warnings with messages that match `filter`. If
    `filter` is not specified all tracer warnings are ignored.

    Note this only installs warning filters if executed within traced code.

    :param filter: A list containing either warning message (str),
        or tuple consisting of (warning message (str), Warning class).
    """
    if not torch._C._get_tracing_state():
        yield
        return

    with warnings.catch_warnings():
        if filter is None:
            warnings.filterwarnings("ignore", category=torch.jit.TracerWarning)
        else:
            for msg in filter:
                category = torch.jit.TracerWarning
                if isinstance(msg, tuple):
                    msg, category = msg
                warnings.filterwarnings("ignore", category=category, message=msg)
        yield


def jit_iter(tensor):
    """
    Iterate over a tensor, ignoring jit warnings.
    """
    # The "Iterating over a tensor" warning is erroneously a RuntimeWarning
    # so we use a custom filter here.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", "Iterating over a tensor")
        warnings.filterwarnings("ignore", category=torch.jit.TracerWarning)
        return list(tensor)


class optional:
    """
    Optionally wrap inside `context_manager` if condition is `True`.
    """

    def __init__(self, context_manager, condition):
        self.context_manager = context_manager
        self.condition = condition

    def __enter__(self):
        if self.condition:
            return self.context_manager.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.condition:
            return self.context_manager.__exit__(exc_type, exc_val, exc_tb)


class ExperimentalWarning(UserWarning):
    pass


@contextmanager
def ignore_experimental_warning():
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=ExperimentalWarning)
        yield


def deep_getattr(obj, name):
    """
    Python getattr() for arbitrarily deep attributes
    Throws an AttributeError if bad attribute
    """
    return functools.reduce(getattr, name.split("."), obj)


class timed:
    def __enter__(self, timer=timeit.default_timer):
        self.start = timer()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end = timeit.default_timer()
        self.elapsed = self.end - self.start
        return self.elapsed


def torch_float(x):
    return x.float() if isinstance(x, torch.Tensor) else float(x)
